
from siliconcompiler.tools.yosys import syn_setup, syn_post_process
import os
import json
import re
from siliconcompiler.tools.yosys.prepareLib import processLibertyFile
from siliconcompiler import sc_open
from siliconcompiler import utils
from siliconcompiler.tools._common.asic import set_tool_task_var, get_libraries, get_mainlib, \
    CellArea
from siliconcompiler.tools._common.asic_clock import get_clock_period
from siliconcompiler.tools._common import get_tool_task


def make_docs(chip):
    from siliconcompiler.targets import asap7_demo
    chip.use(asap7_demo)


def setup(chip):
    '''
    Perform ASIC synthesis
    '''

    # Generic synthesis task setup.
    syn_setup(chip)

    # ASIC-specific setup.
    setup_asic(chip)


def setup_asic(chip):
    ''' Helper method for configs specific to ASIC steps (both syn and lec).
    '''

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['asic', 'logiclib']),
             step=step, index=index)

    syn_corners = get_synthesis_corner(chip)

    if syn_corners is None:
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'synthesis_corner']),
                 step=step, index=index)

    if syn_corners is not None:
        # add timing library requirements
        for lib in get_libraries(chip, 'logic'):
            # mandatory for logiclibs
            chip.add('tool', tool, 'task', task, 'require',
                     ",".join(_get_synthesis_library_key(chip, lib, syn_corners)),
                     step=step, index=index)

        for lib in get_libraries(chip, 'macro'):
            # optional for macrolibs
            if chip.valid(*_get_synthesis_library_key(chip, lib, syn_corners)):
                chip.add('tool', tool, 'task', task, 'require',
                         ",".join(_get_synthesis_library_key(chip, lib, syn_corners)),
                         step=step, index=index)
            elif chip.valid('library', lib, 'output', 'blackbox', 'verilog'):
                chip.add('tool', tool, 'task', task, 'require',
                         ",".join(['library', lib, 'output', 'blackbox', 'verilog']),
                         step=step, index=index)

    if chip.valid('input', 'constraint', 'sdc') and \
       chip.get('input', 'constraint', 'sdc', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,constraint,sdc',
                 step=step, index=index)

    # set default control knobs
    mainlib = get_mainlib(chip)
    for option, value in [
            ('flatten', True),
            ('auto_flatten', True),
            ('hier_threshold', 1000),
            ('autoname', True),
            ('add_buffers', True)]:
        chip.set('tool', tool, 'task', task, 'var', option, value,
                 step=step, index=index, clobber=False)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', option]),
                 step=step, index=index)

    # Add conditionally required mainlib variables
    if chip.valid('library', mainlib, 'option', 'var', 'yosys_buffer_cell'):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'var', 'yosys_buffer_input']),
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'var', 'yosys_buffer_output']),
                 step=step, index=index)

    library_has_addermap = \
        chip.valid('library', mainlib, 'option', 'file', 'yosys_addermap') and \
        chip.get('library', mainlib, 'option', 'file', 'yosys_addermap')
    chip.set('tool', tool, 'task', task, 'var', 'map_adders',
             "true" if library_has_addermap else "false",
             step=step, index=index, clobber=False)
    if chip.get('tool', tool, 'task', task, 'var', 'map_adders', step=step, index=index)[0] == \
       "true":
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'map_adders']),
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'file', 'yosys_addermap']),
                 step=step, index=index)
    library_has_tbufmap = \
        chip.valid('library', mainlib, 'option', 'file', 'yosys_tbufmap') and \
        chip.get('library', mainlib, 'option', 'file', 'yosys_tbufmap')
    if library_has_tbufmap:
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['library', mainlib, 'option', 'file', 'yosys_tbufmap']),
                 step=step, index=index)

    for var0, var1 in [('memory_libmap', 'memory_techmap')]:
        key0 = ['tool', tool, 'tak', task, 'file', var0]
        key1 = ['tool', tool, 'tak', task, 'file', var1]
        if chip.valid(*key0) and chip.get(*key0, step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require', ",".join(key1), step=step, index=index)
        if chip.valid(*key1) and chip.get(*key1, step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require', ",".join(key0), step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'synthesis_corner', get_synthesis_corner(chip),
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'synthesis_corner']),
             step=step, index=index)

    # Require abc clock conversion factor, from library units to ps
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['library', mainlib, 'option', 'var', 'yosys_abc_clock_multiplier']),
             step=step, index=index)
    abc_driver = get_abc_driver(chip)
    if abc_driver:
        chip.set('tool', tool, 'task', task, 'var', 'abc_constraint_driver', abc_driver,
                 step=step, index=index, clobber=False)

    set_tool_task_var(chip, 'abc_constraint_load',
                      schelp='Capacitive load for the abc techmapping in fF, '
                             'if not specified it will not be used',
                      require='lib')

    # document parameters
    chip.set('tool', tool, 'task', task, 'var', 'preserve_modules',
             'List of modules in input files to prevent flatten from "flattening"', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'blackbox_modules',
             'List of modules in input files to exclude from synthesis by replacing them '
             'with empty blackboxes"', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'flatten',
             'true/false, invoke synth with the -flatten option', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'autoname',
             'true/false, call autoname to rename wires based on registers', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'map_adders',
             'true/false, techmap adders in Yosys', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'synthesis_corner',
             'Timing corner to use for synthesis', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'abc_constraint_driver',
             'Buffer that drives the abc techmapping, defaults to first buffer specified',
             field='help')
    chip.set('tool', tool, 'task', task, 'file', 'abc_constraint_file',
             'File used to pass in constraints to abc', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'abc_clock_period',
             'Clock period to use for synthesis in ps, if more than one clock is specified, the '
             'smallest period is used.', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'abc_clock_derating',
             'Used to derate the clock period to further constrain the clock, '
             'values between 0 and 1', field='help')
    chip.set('tool', tool, 'task', task, 'file', 'techmap',
             'File to use for techmapping in Yosys', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'add_buffers',
             'true/false, flag to indicate whether to add buffers or not.', field='help')

    chip.set('tool', tool, 'task', task, 'var', 'auto_flatten',
             'true/false, attempt to determine how to flatten the design', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'hier_iterations',
             'Number of iterations to attempt to determine the hierarchy to flatten',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'hier_threshold',
             'Instance limit for the number of cells in a module to preserve.',
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'hierarchy_separator',
             'control the hierarchy separator used during design flattening', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'hierarchy_separator', '/',
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'hierarchy_separator']),
             step=step, index=index)

    set_tool_task_var(chip, 'map_clockgates',
                      default_value=False,
                      schelp='Map clockgates during synthesis.')

    set_tool_task_var(chip, 'min_clockgate_fanout',
                      default_value=8,
                      schelp='Minimum clockgate fanout.')

    chip.set('tool', tool, 'task', task, 'var', 'strategy',
             'ABC synthesis strategy. Allowed values are DELAY0-4, AREA0-3, or if the strategy '
             'starts with a + it is assumed to be actual commands for ABC.',
             field='help')

    chip.set('tool', tool, 'task', task, 'file', 'memory_libmap',
             'File used to map memories with yosys', field='help')
    chip.set('tool', tool, 'task', task, 'file', 'memory_techmap',
             'File used to techmap memories with yosys', field='help')

    chip.add('tool', tool, 'task', task, 'file', 'synth_extra_map',
             'tools/yosys/techmaps/lcu_kogge_stone.v', package='siliconcompiler',
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'file', 'synth_extra_map',
             'Files used in synthesis to perform additional techmapping', field='help')

    chip.set('tool', tool, 'task', task, 'var', 'lock_design', False,
             step=step, index=index,
             clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'lock_design',
             'true/false, if true will attempt to lock the design with moosic',
             field='help')
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'lock_design']),
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'lock_design_key',
             'lock locking key',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'lock_design_port', 'moosic_key',
             step=step, index=index,
             clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'lock_design_port',
             'lock locking port name',
             field='help')
    if chip.get('tool', tool, 'task', task, 'var', 'lock_design', step=step, index=index)[0] == \
            'true':
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'lock_design_key']),
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'lock_design_port']),
                 step=step, index=index)


################################
# mark cells dont use and format liberty files for yosys and abc
################################
def prepare_synthesis_libraries(chip):
    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    corners = chip.get('tool', tool, 'task', task, 'var', 'synthesis_corner',
                       step=step, index=index)

    logger = None if chip.get('option', 'quiet', step=step, index=index) else chip.logger

    # Clear in case of rerun
    for libtype in ('synthesis_libraries', 'synthesis_libraries_macros'):
        chip.set('tool', tool, 'task', task, 'file', libtype, [],
                 step=step, index=index)
        chip.set('tool', tool, 'task', task, 'file', libtype, False, field='copy')

    # Generate synthesis_libraries and synthesis_macro_libraries for Yosys use

    # mark libs with dont_use since ABC cannot get this information via its commands
    # this also ensures the liberty files have been decompressed and corrected formatting
    # issues that generally cannot be handled by yosys or yosys-abc
    def get_synthesis_libraries(lib):
        keypath = _get_synthesis_library_key(chip, lib, corners)
        if chip.valid(*keypath):
            return chip.find_files(*keypath, step=step, index=index)
        return []

    lib_file_map = {}
    for libtype in ('logic', 'macro'):
        for lib in get_libraries(chip, libtype):
            lib_content = {}
            lib_map = {}
            # Mark dont use
            for lib_file in get_synthesis_libraries(lib):
                # Ensure a unique name is used for library
                lib_file_name_base = os.path.basename(lib_file)
                if lib_file_name_base.lower().endswith('.gz'):
                    lib_file_name_base = lib_file_name_base[0:-3]
                if lib_file_name_base.lower().endswith('.lib'):
                    lib_file_name_base = lib_file_name_base[0:-4]

                lib_file_name = lib_file_name_base
                unique_ident = 0
                while lib_file_name in lib_content:
                    lib_file_name = f'{lib_file_name_base}_{unique_ident}'
                    unique_ident += 1

                lib_content[lib_file_name] = processLibertyFile(
                    lib_file,
                    logger=logger
                )
                lib_map[lib_file_name] = lib_file

            if not lib_content:
                continue

            var_name = 'synthesis_libraries'
            if libtype == "macro":
                var_name = 'synthesis_libraries_macros'

            for file, content in lib_content.items():
                output_file = os.path.join(
                    chip.getworkdir(step=step, index=index),
                    'inputs',
                    f'sc_{libtype}_{lib}_{file}.lib'
                )
                lib_file_map[lib_map[file]] = output_file

                with open(output_file, 'w') as f:
                    f.write(content)

                chip.add('tool', tool, 'task', task, 'file', var_name, output_file,
                         step=step, index=index)


def create_abc_synthesis_constraints(chip):

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    abc_driver = chip.get('tool', tool, 'task', task, 'var', 'abc_constraint_driver',
                          step=step, index=index)
    if abc_driver:
        abc_driver = abc_driver[0]

    abc_load = chip.get('tool', tool, 'task', task, 'var', 'abc_constraint_load',
                        step=step, index=index)
    if abc_load:
        abc_load = abc_load[0]

    if not abc_driver and not abc_load:
        # neither is set so nothing to do
        return

    with open(chip.get('tool', tool, 'task', task, 'file', 'abc_constraint_file',
                       step=step, index=index)[0], "w") as f:
        if abc_load:
            # convert to fF
            abc_load = re.match(r'([0-9]*\.?[0-9]*)\s*([fpa])[fF]?', abc_load)

            if not abc_load:
                raise ValueError(f'Unable to parse {abc_load}')

            abc_load_unit = abc_load.group(2)
            if not abc_load_unit:
                abc_load_unit = chip.get('unit', 'capacitance')[0]

            abc_load = float(abc_load.group(1))

            if abc_load_unit == 'p':
                abc_load *= 1000
            elif abc_load_unit == 'a':
                abc_load /= 1000

        abc_template = utils.get_file_template('abc.const',
                                               root=os.path.join(os.path.dirname(__file__),
                                                                 'templates'))
        f.write(abc_template.render(abc_driver=abc_driver, abc_load=abc_load))


def get_synthesis_corner(chip):
    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    syn_corners = chip.get('tool', tool, 'task', task, 'var', 'synthesis_corner',
                           step=step, index=index)
    if syn_corners:
        return syn_corners

    # determine corner based on setup corner from constraints
    corner = None
    for constraint in chip.getkeys('constraint', 'timing'):
        checks = chip.get('constraint', 'timing', constraint, 'check', step=step, index=index)
        if "setup" in checks and not corner:
            corner = chip.get('constraint', 'timing', constraint, 'libcorner',
                              step=step, index=index)

    if not corner:
        # try getting it from first constraint with a valid libcorner
        for constraint in chip.getkeys('constraint', 'timing'):
            if not corner:
                corner = chip.get('constraint', 'timing', constraint, 'libcorner',
                                  step=step, index=index)

    return corner


def _get_synthesis_library_key(chip, lib, corners):
    if chip.valid('library', lib, 'option', 'file', 'yosys_synthesis_libraries'):
        return ('library', lib, 'option', 'file', 'yosys_synthesis_libraries')

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    delaymodel = chip.get('asic', 'delaymodel', step=step, index=index)
    for corner in corners:
        if chip.valid('library', lib, 'output', corner, delaymodel):
            return ('library', lib, 'output', corner, delaymodel)

    return ('library', lib, 'output', corners[0], delaymodel)


def get_abc_period(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    mainlib = get_mainlib(chip)

    abc_clock_period = chip.get('tool', tool, 'task', task, 'var', 'abc_clock_period',
                                step=step, index=index)
    if abc_clock_period:
        return abc_clock_period[0]

    abc_clock_multiplier = float(chip.get('library', mainlib, 'option', 'var',
                                          'yosys_abc_clock_multiplier')[0])

    _, period = get_clock_period(chip,
                                 clock_units_multiplier=abc_clock_multiplier / 1000)
    if not period:
        return None

    period *= 1000

    abc_clock_derating = chip.get('tool', tool, 'task', task, 'var', 'abc_clock_derating',
                                  step=step, index=index)
    if abc_clock_derating:
        derating = float(abc_clock_derating[0])
        if derating > 1:
            chip.logger.warning("abc_clock_derating is greater than 1.0")
        elif derating > 0:
            period *= (1.0 - derating)
        else:
            chip.logger.error("abc_clock_derating is negative")

    return int(period)


def get_abc_driver(chip):

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    abc_driver = chip.get('tool', tool, 'task', task, 'var', 'abc_constraint_driver',
                          step=step, index=index)
    if abc_driver:
        return abc_driver[0]

    abc_driver = None
    # get the first driver defined in the logic lib
    for lib in get_libraries(chip, 'logic'):
        if chip.valid('library', lib, 'option', 'var', 'yosys_driver_cell') and not abc_driver:
            abc_driver = chip.get('library', lib, 'option', 'var', 'yosys_driver_cell')[0]

    return abc_driver


##################################################
def pre_process(chip):
    ''' Tool specific function to run before step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    # copy techmapping from libraries
    logiclibs = get_libraries(chip, 'logic')
    macrolibs = get_libraries(chip, 'macro')
    for lib in logiclibs + macrolibs:
        if not chip.valid('library', lib, 'option', 'file', 'yosys_techmap'):
            continue
        for techmap in chip.find_files('library', lib, 'option', 'file', 'yosys_techmap'):
            if techmap is None:
                continue
            chip.add('tool', tool, 'task', task, 'file', 'techmap', techmap, step=step, index=index)

    # Constants needed by yosys, do not allow overriding of values so force clobbering
    chip.set('tool', tool, 'task', task, 'file', 'abc_constraint_file',
             f"{chip.getworkdir(step=step, index=index)}/inputs/sc_abc.constraints",
             step=step, index=index, clobber=True)
    chip.set('tool', tool, 'task', task, 'file', 'abc_constraint_file', False, field='copy')

    abc_clock_period = get_abc_period(chip)
    if abc_clock_period:
        chip.set('tool', tool, 'task', task, 'var', 'abc_clock_period', str(abc_clock_period),
                 step=step, index=index, clobber=False)

    prepare_synthesis_libraries(chip)
    create_abc_synthesis_constraints(chip)


def post_process(chip):
    syn_post_process(chip)
    _generate_cell_area_report(chip)


def _generate_cell_area_report(chip):
    design = "gcd"
    if not os.path.exists('reports/stat.json'):
        return
    if not os.path.exists(f'outputs/{design}.netlist.json'):
        return

    with sc_open('reports/stat.json') as fd:
        stat = json.load(fd)

    with sc_open(f'outputs/{design}.netlist.json') as fd:
        netlist = json.load(fd)
    modules = []
    for module in stat["modules"].keys():
        if module[0] == "\\":
            modules.append(module[1:])

    cellarea_report = CellArea()

    def get_area_count(module):
        if f"\\{module}" not in stat["modules"]:
            return 0.0, 0
        info = stat["modules"][f"\\{module}"]

        count = info["num_cells"]
        area = 0.0
        if "area" in info:
            area = info["area"]

        for cell, inst_count in info["num_cells_by_type"].items():
            # print(module, cell, inst_count)

            cell_area, cell_count = get_area_count(cell)

            count += cell_count * inst_count
            if cell_count > 0:
                count -= inst_count
            area += cell_area * inst_count

        return area, count

    def handle_heir(level_info, prefix):
        cells = list(level_info["cells"])

        for cell in cells:
            cell_type = level_info["cells"][cell]["type"]
            if cell_type in modules:
                area, count = get_area_count(cell_type)
                cellarea_report.addCell(
                    name=f"{prefix}{cell}",
                    module=cell_type,
                    cellcount=count,
                    cellarea=area)
                handle_heir(netlist["modules"][cell_type], f"{prefix}{cell}.")

    count = stat["design"]["num_cells"]
    area = 0.0
    if "area" in stat["design"]:
        area = stat["design"]["area"]
    cellarea_report.addCell(
        name=design,
        module=design,
        cellarea=area,
        cellcount=count
    )

    handle_heir(netlist["modules"][design], "")

    if cellarea_report.size() > 0:
        cellarea_report.writeReport("reports/hierarchical_cell_area.json")



import siliconcompiler
from siliconcompiler.flows import asicflow, asictopflow, signoffflow, synflow
from siliconcompiler.checklists import oh_tapeout
from siliconcompiler.tools import openroad
from siliconcompiler.tools._common import get_tool_tasks

from lambdapdk import gf180
from lambdapdk.gf180.libs import gf180mcu


####################################################
# Target Setup
####################################################
def setup(chip, syn_np=1, floorplan_np=1, physyn_np=1, place_np=1, cts_np=1, route_np=1,
          timing_np=1):
    '''
    Global foundries 180 Demo Target
    '''

    asic_flow_args = {
        "syn_np": syn_np,
        "floorplan_np": floorplan_np,
        "physyn_np": physyn_np,
        "place_np": place_np,
        "cts_np": cts_np,
        "route_np": route_np
    }

    # 1. Load PDK, flow, libs
    chip.use(gf180)
    chip.use(gf180mcu)
    chip.use(asicflow, **asic_flow_args)
    chip.use(synflow, syn_np=syn_np, timing_np=timing_np)
    chip.use(asictopflow)
    chip.use(signoffflow)
    chip.use(oh_tapeout)

    # 2. Set default targets
    chip.set('option', 'flow', 'asicflow', clobber=False)
    chip.set('option', 'pdk', 'gf180', clobber=False)
    chip.set('option', 'stackup', '5LM_1TM_9K', clobber=False)

    # 3. Set project specific design choices
    chip.set('asic', 'logiclib', 'gf180mcu_fd_sc_mcu9t5v0', clobber=False)

    # 4. get project specific design choices
    chip.set('asic', 'delaymodel', 'nldm', clobber=False)
    chip.set('constraint', 'density', 40, clobber=False)
    chip.set('constraint', 'coremargin', 1, clobber=False)

    # 5. Timing corners
    chip.set('constraint', 'timing', 'slow', 'libcorner', 'slow', clobber=False)
    chip.set('constraint', 'timing', 'slow', 'pexcorner', 'wst', clobber=False)
    chip.set('constraint', 'timing', 'slow', 'mode', 'func', clobber=False)
    chip.set('constraint', 'timing', 'slow', 'check', ['setup', 'hold'], clobber=False)

    chip.set('constraint', 'timing', 'fast', 'libcorner', 'fast', clobber=False)
    chip.set('constraint', 'timing', 'fast', 'pexcorner', 'bst', clobber=False)
    chip.set('constraint', 'timing', 'fast', 'mode', 'func', clobber=False)
    chip.set('constraint', 'timing', 'fast', 'check', ['setup', 'hold'], clobber=False)

    chip.set('constraint', 'timing', 'typical', 'libcorner', 'typical', clobber=False)
    chip.set('constraint', 'timing', 'typical', 'pexcorner', 'typ', clobber=False)
    chip.set('constraint', 'timing', 'typical', 'mode', 'func', clobber=False)
    chip.set('constraint', 'timing', 'typical', 'check', ['power'], clobber=False)

    # PSM gets stuck in a loop, must be disabled for now on gf180
    for task in get_tool_tasks(chip, openroad):
        chip.set('tool', 'openroad', 'task', task, 'var', 'psm_enable', False)
    chip.set('tool', 'openroad', 'task', 'route', 'var', 'ant_check', False)


#########################
if __name__ == "__main__":
    target = siliconcompiler.Chip('<target>')
    setup(target)
    target.write_manifest('gf180_demo.json')



# KLayout script to export a .GDS file from a .DEF-formatted design.
#
# Source: The OpenROAD Project.
# https://github.com/The-OpenROAD-Project/OpenROAD-flow-scripts/blob/master/flow/util/def2stream.py
#
# License: BSD 3-Clause.
#
# Copyright (c) 2018, The Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pya
import os
import sys
import fnmatch


def gds_export(design_name, in_def, in_files, out_file, tech, allow_missing, config_file='',
               seal_file='',
               timestamps=True):
    from klayout_utils import get_write_options  # noqa E402

    # Load def file
    main_layout = pya.Layout()
    main_layout.technology_name = tech.name
    main_layout.read(in_def, tech.load_layout_options)

    # List cells
    def_cells = []
    for def_cell in main_layout.each_cell():
        def_cells.append(def_cell.name)

    def_cells.remove(design_name)
    # Remove vias
    def_cells = sorted([cell for cell in def_cells if not cell.startswith("VIA_")])
    print(f"[INFO] Read in {len(def_cells)} cells from DEF file")
    for cell in def_cells:
        print(f"  [INFO] DEF cell: {cell}")

    if f"{design_name}_DEF_FILL" in def_cells:
        def_cells.remove(f"{design_name}_DEF_FILL")

    # Load in the gds to merge
    print("[INFO] Merging GDS/OAS files...")
    for fil in in_files:
        macro_layout = pya.Layout()
        macro_layout.read(fil)
        print(f"[INFO] Read in {fil}")
        for cell in list(def_cells):
            if macro_layout.has_cell(cell):
                subcell = main_layout.cell(cell)
                print(f"  [INFO] Merging in {cell}")
                subcell.copy_tree(macro_layout.cell(cell))
                def_cells.remove(cell)

    # Copy the top level only to a new layout
    print("[INFO] Copying toplevel cell '{0}'".format(design_name))
    top_only_layout = pya.Layout()
    top_only_layout.dbu = main_layout.dbu
    top = top_only_layout.create_cell(design_name)
    top.copy_tree(main_layout.cell(design_name))

    print("[INFO] Checking for missing GDS/OAS...")
    missing_cell = False
    for check_cell in def_cells:
        missing_cell = True
        allowed_missing = any([fnmatch.fnmatch(check_cell, pattern) for pattern in allow_missing])
        print(f"[{'WARNING' if allowed_missing else 'ERROR'}] LEF Cell '{check_cell}' has no "
              "matching GDS/OAS cell. Cell will be empty")

    if not missing_cell:
        print("[INFO] All LEF cells have matching GDS/OAS cells")

    print("[INFO] Checking for orphan cell in the final layout...")
    for i in top_only_layout.each_cell():
        if i.name != design_name and i.parent_cells() == 0:
            print("[ERROR] Found orphan cell '{0}'".format(i.name))

    if seal_file:
        top_cell = top_only_layout.top_cell()

        print("[INFO] Reading seal GDS/OAS file...")
        print("\t{0}".format(seal_file))
        top_only_layout.read(seal_file)

        for cell in top_only_layout.top_cells():
            if cell != top_cell:
                print("[INFO] Merging '{0}' as child of '{1}'".format(cell.name, top_cell.name))
                top.insert(pya.CellInstArray(cell.cell_index(), pya.Trans()))

    # Write out the GDS
    print("[INFO] Writing out GDS/OAS '{0}'".format(out_file))
    top_only_layout.write(out_file, get_write_options(out_file, timestamps))


def main():
    # SC_ROOT provided by CLI
    sys.path.append(SC_KLAYOUT_ROOT)  # noqa: F821
    sys.path.append(SC_TOOLS_ROOT)  # noqa: F821
    print(sys.path)

    from klayout_utils import (
        technology,
        get_streams,
        save_technology,
        get_schema,
        generate_metrics
    )
    from klayout_show import show
    from _common.asic import get_libraries

    schema = get_schema(manifest='sc_manifest.json')

    # Extract info from manifest
    sc_step = schema.get('arg', 'step')
    sc_index = schema.get('arg', 'index')
    sc_pdk = schema.get('option', 'pdk')
    sc_flow = schema.get('option', 'flow')
    sc_task = schema.get('flowgraph', sc_flow, sc_step, sc_index, 'task')
    sc_klayout_vars = schema.getkeys('tool', 'klayout', 'task', sc_task, 'var')
    sc_stream = schema.get('tool', 'klayout', 'task', sc_task, 'var', 'stream',
                           step=sc_step, index=sc_index)[0]

    if schema.valid('option', 'stackup'):
        sc_stackup = schema.get('option', 'stackup')
    else:
        sc_stackup = schema.get('pdk', sc_pdk, 'stackup')[0]

    design = schema.get('option', 'entrypoint')
    if not design:
        design = schema.get('design')

    in_def = None
    for ext in ('def.gz', 'def'):
        in_def = os.path.join('inputs', f'{design}.{ext}')
        if os.path.exists(in_def):
            break
        in_def = None
    if not in_def:
        in_def = schema.get('input', 'layout', 'def', step=sc_step, index=sc_index)[0]

    out_file = os.path.join('outputs', f'{design}.{sc_stream}')

    libs = get_libraries(schema, 'logic')
    libs += get_libraries(schema, 'macro')

    in_files = []
    for lib in libs:
        for s in get_streams(schema):
            if schema.valid('library', lib, 'output', sc_stackup, s):
                in_files.extend(schema.get('library', lib, 'output', sc_stackup, s,
                                           step=sc_step, index=sc_index))
                break

    allow_missing = []
    for lib in libs:
        if schema.valid('library', lib, 'option', 'var', 'klayout_allow_missing_cell'):
            patterns = [pattern for pattern in schema.get('library', lib, 'option', 'var',
                                                          'klayout_allow_missing_cell') if pattern]
            allow_missing.extend(patterns)

    if 'timestamps' in sc_klayout_vars:
        sc_timestamps = schema.get('tool', 'klayout', 'task', sc_task, 'var', 'timestamps',
                                   step=sc_step, index=sc_index) == ['true']
    else:
        sc_timestamps = False

    if 'screenshot' in sc_klayout_vars:
        sc_screenshot = schema.get('tool', 'klayout', 'task', sc_task, 'var', 'screenshot',
                                   step=sc_step, index=sc_index) == ['true']
    else:
        sc_screenshot = True

    sc_tech = technology(design, schema)

    gds_export(design, in_def, in_files, out_file, sc_tech, allow_missing,
               config_file='', seal_file='', timestamps=sc_timestamps)

    if sc_screenshot:
        show(schema, sc_tech, out_file, f'outputs/{design}.png', screenshot=True)

    # Save tech files
    save_technology(design, sc_tech)

    generate_metrics()


if __name__ == '__main__':
    main()



# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.
import os
import sys

from siliconcompiler import Chip
from siliconcompiler import SiliconCompilerError
from siliconcompiler.remote.client import Client, ConfigureClient
from siliconcompiler.scheduler import _finalize_run
from siliconcompiler.flowgraph import _get_flowgraph_entry_nodes, _get_flowgraph_node_outputs, \
    nodes_to_execute


def main():
    progname = "sc-remote"
    description = """
-----------------------------------------------------------
SC app that provides an entry point to common remote / server
interactions.

To generate a configuration file, use:
    sc-remote -configure

    or to specify a specific server and/or port:
    sc-remote -configure -server https://example.com
    sc-remote -configure -server https://example.com:1234

    to add or remove directories from upload whitelist, these
        also support globbing:
    sc-remote -configure -add ./fine_to_upload
    sc-remote -configure -remove ./no_longer_okay_to_upload

    to display the full configuration of the credentials file
    sc-remote -configure -list

To check an ongoing job's progress, use:
    sc-remote -cfg <stepdir>/outputs/<design>.pkg.json

To cancel an ongoing job, use:
    sc-remote -cancel -cfg <stepdir>/outputs/<design>.pkg.json

To reconnect an ongoing job, use:
    sc-remote -reconnect -cfg <stepdir>/outputs/<design>.pkg.json

To delete a job, use:
    sc-remote -delete -cfg <stepdir>/outputs/<design>.pkg.json
-----------------------------------------------------------
"""

    # Argument Parser
    progname = 'sc-remote'
    chip = Chip(progname)
    switchlist = ['-cfg', '-credentials']
    extra_args = {
        '-configure': {'action': 'store_true',
                       'help': 'create configuration file for the remote',
                       'sc_print': False},
        '-server': {'help': 'address of server for configure (only valid with -configure)',
                    'metavar': '<server>',
                    'sc_print': False},
        '-add': {'help': 'path to add to the upload whitelist (only valid with -configure)',
                 'metavar': '<path>',
                 'nargs': '+',
                 'sc_print': False},
        '-remove': {'help': 'path to remove from the upload whitelist (only valid with -configure)',
                    'metavar': '<path>',
                    'nargs': '+',
                    'sc_print': False},
        '-list': {'help': 'print the current configuration (only valid with -configure)',
                  'action': 'store_true',
                  'sc_print': False},
        '-reconnect': {'action': 'store_true',
                       'help': 'reconnect to a running job on the remote',
                       'sc_print': False},
        '-cancel': {'action': 'store_true',
                    'help': 'cancel a running job on the remote',
                    'sc_print': False},
        '-delete': {'action': 'store_true',
                    'help': 'delete a job on the remote',
                    'sc_print': False},
    }

    try:
        args = chip.create_cmdline(progname,
                                   switchlist=switchlist,
                                   additional_args=extra_args,
                                   description=description)
    except Exception as e:
        chip.logger.error(e)
        return 1

    # Sanity checks.
    exclusive = ['configure', 'reconnect', 'cancel', 'delete']
    cfg_only = ['reconnect', 'cancel', 'delete']

    exclusive_count = sum([1 for arg in exclusive if args[arg]])
    if exclusive_count > 1:
        chip.logger.error(f'Error: {", ".join(["-"+e for e in exclusive])} are mutually exclusive')
        return 1
    chip_cfg = chip.get('option', 'cfg')
    if chip_cfg and not any([args[arg] for arg in cfg_only]):
        chip.logger.error(f'Error: -cfg is required for {", ".join(["-"+e for e in cfg_only])}')
    if any([args[arg] for arg in cfg_only]) and args['server']:
        chip.logger.error('Error: -server cannot be specified with '
                          f'{", ".join(["-"+e for e in cfg_only])}')

    if args['configure']:
        if args['list']:
            client = Client(chip)
            client.print_configuration()
            return 0

        if not args['add'] and not args['remove']:
            try:
                client = ConfigureClient(chip)
                client.configure_server(server=args['server'])
            except ValueError as e:
                chip.logger.error(e)
                return 1
        else:
            try:
                client = ConfigureClient(chip)
                client.configure_whitelist(add=args['add'], remove=args['remove'])
            except ValueError as e:
                chip.logger.error(e)
                return 1

        return 0

    client = Client(chip)
    # Main logic.
    # If no job-related options are specified, fetch and report basic info.
    # Create temporary Chip object and check on the server.
    try:
        client.check()
    except SiliconCompilerError as e:
        chip.logger.error(f'{e}')
        return 1

    # If the -cancel flag is specified, cancel the job.
    if args['cancel']:
        try:
            client.cancel_job()
        except SiliconCompilerError as e:
            chip.logger.error(f'{e}')
            return 1

    # If the -delete flag is specified, delete the job.
    elif args['delete']:
        try:
            client.delete_job()
        except SiliconCompilerError as e:
            chip.logger.error(f'{e}')
            return 1

    # If the -reconnect flag is specified, re-enter the client flow
    # in its "check_progress/ until job is done" loop.
    elif args['reconnect']:
        # Start from successors of entry nodes, so entry nodes are not fetched from remote.
        flow = chip.get('option', 'flow')
        entry_nodes = _get_flowgraph_entry_nodes(chip, flow)
        for entry_node in entry_nodes:
            outputs = _get_flowgraph_node_outputs(chip, flow, entry_node)
            chip.set('option', 'from', list(map(lambda node: node[0], outputs)))
        # Enter the remote run loop.
        try:
            client._run_loop()
        except SiliconCompilerError as e:
            chip.logger.error(f'{e}')
            return 1

        # Wrap up run
        for step, index in nodes_to_execute(chip):
            manifest = os.path.join(chip.getworkdir(step=step, index=index),
                                    'outputs',
                                    f'{chip.design}.pkg.json')
            if os.path.exists(manifest):
                chip.schema.read_journal(manifest)
        _finalize_run(chip)

        # Summarize the run.
        chip.summary()

    # If only a manifest is specified, make a 'check_progress/' request and report results:
    elif chip_cfg:
        try:
            info = client.check_job_status()
            client._report_job_status(info)
        except Exception as e:
            chip.logger.error(f'{e}')
            return 1

    # Done
    return 0


#########################
if __name__ == "__main__":
    sys.exit(main())



'''
Verilator is a free and open-source software tool which converts Verilog (a
hardware description language) to a cycle-accurate behavioral model in C++
or SystemC.

All Verilator tasks may consume input either from a single pickled Verilog file
(``inputs/<design>.v``) generated by a preceding task, or if that file does not
exist, through the following keypaths:

    * :keypath:`input, rtl, verilog`
    * :keypath:`input, cmdfile, f`
    * :keypath:`option, ydir`
    * :keypath:`option, vlib`
    * :keypath:`option, idir`

For all tasks, this driver runs Verilator using the ``-sv`` switch to enable
parsing a subset of SystemVerilog features.

Documentation: https://verilator.org/guide/latest

Sources: https://github.com/verilator/verilator

Installation: https://verilator.org/guide/latest/install.html
'''

import os
from siliconcompiler import utils
from siliconcompiler.tools._common import (
    add_frontend_requires,
    get_frontend_options,
    add_require_input,
    get_input_files,
    get_tool_task,
    input_provides
)


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    from siliconcompiler.targets import freepdk45_demo
    chip.use(freepdk45_demo)


def setup(chip):
    ''' Per tool function that returns a dynamic options string based on
    the dictionary settings. Static settings only.
    '''

    tool = 'verilator'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Basic Tool Setup
    chip.set('tool', tool, 'exe', 'verilator')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=4.034', clobber=False)

    # Common to all tasks
    # Max threads
    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=False)

    # Basic warning and error grep check on logfile
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', r"^\%Warning",
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', r"^\%Error",
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'file', 'config',
             'Verilator configuration file',
             field='help')
    add_require_input(chip, 'tool', tool, 'task', task, 'file', 'config')
    add_require_input(chip, 'option', 'file', 'verilator_config')

    chip.set('tool', tool, 'task', task, 'var', 'enable_assert',
             'true/false, when true assertions are enabled in Verilator.',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'enable_assert', False,
             step=step, index=index, clobber=False)

    if chip.get('tool', tool, 'task', task, 'var', 'enable_assert', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', 'enable_assert']),
                 step=step, index=index)

    if f'{chip.top()}.v' not in input_provides(chip, step, index):
        add_require_input(chip, 'input', 'rtl', 'verilog')
        add_require_input(chip, 'input', 'rtl', 'systemverilog')
    add_require_input(chip, 'input', 'cmdfile', 'f')
    add_frontend_requires(chip, ['ydir', 'vlib', 'idir', 'libext', 'param', 'define'])


def runtime_options(chip):
    cmdlist = []
    tool = 'verilator'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    design = chip.top()

    has_input = os.path.isfile(f'inputs/{design}.v')
    opts_supports = ['param', 'libext']
    if not has_input:
        opts_supports.extend(['ydir', 'vlib', 'idir', 'define'])

    frontend_opts = get_frontend_options(chip, opts_supports)

    # Even though most of these don't need to be set in runtime_options() in order for the driver to
    # function properly, setting all the CLI options here facilitates a user using ['tool', <tool>,
    # 'task', <task>, 'option'] to supply additional CLI flags.

    cmdlist.append('-sv')
    cmdlist.extend(['--top-module', design])

    assertions = chip.get('tool', tool, 'task', task, 'var',
                          'enable_assert', step=step, index=index)
    if assertions == ['true']:
        cmdlist.append('--assert')

    # Converting user setting to verilator specific filter
    for warning in chip.get('tool', tool, 'task', task, 'warningoff', step=step, index=index):
        cmdlist.append(f'-Wno-{warning}')

    libext = frontend_opts['libext']
    if libext:
        libext_option = f"+libext+.{'+.'.join(libext)}"
        cmdlist.append(libext_option)

    # Verilator docs recommend this file comes first in CLI arguments
    for value in get_input_files(chip, 'tool', tool, 'task', task, 'file', 'config'):
        cmdlist.append(value)
    for value in get_input_files(chip, 'option', 'file', 'verilator_config'):
        cmdlist.append(value)

    for param, value in frontend_opts['param']:
        value = value.replace('"', '\\"')
        cmdlist.append(f'-G{param}={value}')

    if os.path.isfile(f'inputs/{design}.v'):
        cmdlist.append(f'inputs/{design}.v')
    else:
        for value in frontend_opts['ydir']:
            cmdlist.append(f'-y {value}')
        for value in frontend_opts['vlib']:
            cmdlist.append(f'-v {value}')
        for value in frontend_opts['idir']:
            cmdlist.append(f'-I{value}')
        for value in frontend_opts['define']:
            if value == "VERILATOR":
                # Verilator auto defines this and will error if it is defined twice.
                continue
            cmdlist.append(f'-D{value}')
        for value in get_input_files(chip, 'input', 'rtl', 'systemverilog'):
            cmdlist.append(value)
        for value in get_input_files(chip, 'input', 'rtl', 'verilog'):
            cmdlist.append(value)

    for value in get_input_files(chip, 'input', 'cmdfile', 'f'):
        cmdlist.append(f'-f {value}')

    return cmdlist


################################
# Version Check
################################


def parse_version(stdout):
    # Verilator 4.104 2020-11-14 rev v4.104
    return stdout.split()[1]


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("verilator.csv")



import os
import shlex

from siliconcompiler.tools._common import input_provides, has_input_files, \
    get_input_files, get_tool_task, record_metric
from siliconcompiler.tools._common.asic import set_tool_task_var, get_tool_task_var

from siliconcompiler.tools.klayout.klayout import setup as setup_tool
import xml.etree.ElementTree as ET
from siliconcompiler import utils


def make_docs(chip):
    from siliconcompiler.tools.klayout import klayout
    klayout.make_docs(chip)
    chip.set('tool', 'klayout', 'task', 'drc', 'var', 'drc_name', '<drc_name>',
             step='<step>', index='<index>')


def setup(chip):
    '''
    Performs a design rule check on the provided layout
    '''

    # Generic tool setup.
    setup_tool(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    design = chip.top()

    clobber = False

    option = ['-z', '-nc', '-rx']
    chip.set('tool', tool, 'task', task, 'option', option,
             step=step, index=index, clobber=clobber)

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=clobber)

    chip.add('tool', tool, 'task', task, 'require', 'option,pdk')
    chip.add('tool', tool, 'task', task, 'require', 'option,stackup')

    chip.set('tool', tool, 'task', task, 'var', 'drc_name', 'drc',
             step=step, index=index, clobber=False)
    chip.add('tool', tool, 'task', task, 'require', f'tool,{tool},task,{task},var,drc_name',
             step=step, index=index)

    drc_name = chip.get('tool', tool, 'task', task, 'var', 'drc_name', step=step, index=index)
    if not drc_name:
        raise ValueError('drc_name is required')
    drc_name = drc_name[0]

    pdk = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')
    chip.add('tool', tool, 'task', task, 'require',
             f'pdk,{pdk},drc,runset,klayout,{stackup},{drc_name}')

    if f'{design}.gds' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.gds',
                 step=step, index=index)
    elif f'{design}.oas' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.oas',
                 step=step, index=index)
    elif has_input_files(chip, 'input', 'layout', 'oas', check_library_files=False):
        chip.add('tool', tool, 'task', task, 'require', 'input,layout,oas',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,layout,gds',
                 step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', design + '.lyrdb',
             step=step, index=index)

    set_tool_task_var(
        chip,
        f'drc_params:{drc_name}',
        schelp="Input parameter to DRC script, in the form of key=value, if the value "
               "is <topcell>, <input>, <report>, <threads> these will be automatically "
               "determined.",
        skip='lib')


def runtime_options(chip):
    design = chip.top()

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    pdk = chip.get('option', 'pdk')
    stackup = chip.get('option', 'stackup')

    layout = None
    for file in [f'inputs/{design}.gds', f'inputs/{design}.oas']:
        if os.path.isfile(file):
            layout = file
            break

    if not layout:
        for file in [
                get_input_files(chip, 'input', 'layout', 'oas', add_library_files=False),
                get_input_files(chip, 'input', 'layout', 'gds', add_library_files=False)]:
            if file:
                layout = file[0]

    threads = chip.get('tool', tool, 'task', task, 'threads', step=step, index=index)
    if not threads:
        threads = 1

    drc_name = chip.get('tool', tool, 'task', task, 'var', 'drc_name',
                        step=step, index=index)[0]
    report = os.path.abspath(f"outputs/{chip.top()}.lyrdb")

    runset = chip.find_files('pdk', pdk, 'drc', 'runset', 'klayout', stackup, drc_name)[0]

    params_lookup = {
        "<topcell>": chip.top(),
        "<report>": shlex.quote(report),
        "<threads>": threads,
        "<input>": shlex.quote(layout)
    }

    args = [
        '-r', shlex.quote(runset)
    ]

    for param in get_tool_task_var(chip, f'drc_params:{drc_name}', skip='lib'):
        for lookup, value in params_lookup.items():
            param = param.replace(lookup, str(value))
        args.extend(
            ['-rd', param]
        )
    return args


def post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    drc_db = f"outputs/{chip.top()}.lyrdb"

    drc_report = None
    if os.path.isfile(drc_db):
        with open(drc_db, "r") as f:
            drc_report = ET.fromstring(f.read())
    if drc_report is None:
        drc_db = []

    violation_count = 0
    if drc_report:
        violations = drc_report.find('items')
        if violations:
            violation_count = len(violations.findall('item'))

    record_metric(chip, step, index, 'drcs', violation_count, drc_db)



from siliconcompiler import SiliconCompilerError
from siliconcompiler.tools.vpr import show
from siliconcompiler.tools._common import get_tool_task


def setup(chip, clobber=True):
    '''
    Screenshot placed and/or routed designs
    '''

    show.setup(chip, clobber=clobber)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()
    chip.add('tool', tool, 'task', task, 'output', f'{design}.png', step=step, index=index)


def runtime_options(chip):
    '''Command line options to vpr for screenshot
    '''

    design = chip.top()

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if chip.valid('tool', tool, 'task', task, 'var', 'show_filepath'):
        show_type = chip.get('tool', tool, 'task', task, 'var', 'show_filetype',
                             step=step, index=index)[0]
    else:
        raise SiliconCompilerError("Invalid filepath", chip=chip)

    options = show.generic_show_options(chip)

    if show_type == 'route':
        screenshot_command_str = ("set_draw_block_text 0; " +
                                  "set_draw_block_outlines 0; " +
                                  "set_routing_util 1; " +
                                  f"save_graphics outputs/{design}.png;")
    elif show_type == 'place':
        screenshot_command_str = ("set_draw_block_text 1; " +
                                  "set_draw_block_outlines 1; " +
                                  f"save_graphics outputs/{design}.png;")
    else:
        raise SiliconCompilerError(f"Incorrect file type {show_type}", chip=chip)

    options.append("--graphics_commands")
    options.append(f"\"{screenshot_command_str}\"")

    return options



from siliconcompiler.utils import default_email_credentials_file, get_file_template
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
import json
import os
from siliconcompiler import sc_open
from siliconcompiler.schema import Schema
from siliconcompiler.report import utils as report_utils
import fastjsonschema
from pathlib import Path
from siliconcompiler.flowgraph import get_executed_nodes
import uuid


# Compile validation code for API request bodies.
api_dir = Path(__file__).parent / 'validation'

# 'remote_run': Run a stage of a job using the server's cluster settings.
with open(api_dir / 'email_credentials.json') as schema:
    validate_creds = fastjsonschema.compile(json.loads(schema.read()))


def __load_config(chip):
    path = default_email_credentials_file()
    if not os.path.exists(path):
        chip.logger.warn(f'Email credentials are not available: {path}')
        return {}

    with open(path) as f:
        creds = json.load(f)

    try:
        return validate_creds(creds)
    except fastjsonschema.JsonSchemaException as e:
        chip.logger.error(f'Email credentials failed to validate: {e}')
        return {}


def send(chip, msg_type, step, index):
    chip_step, chip_index = step, index
    if step is None:
        chip_step = Schema.GLOBAL_KEY
    if index is None:
        chip_index = Schema.GLOBAL_KEY
    to = chip.get('option', 'scheduler', 'msgcontact', step=chip_step, index=chip_index)
    event = chip.get('option', 'scheduler', 'msgevent', step=chip_step, index=chip_index)

    if not to or not event:
        # nothing to do
        return

    if 'all' not in event and msg_type not in event:
        # nothing to do
        return

    cred = __load_config(chip)

    if not cred:
        return

    jobname = chip.get("option", "jobname")
    flow = chip.get("option", "flow")

    msg = MIMEMultipart()

    if step and index:
        subject = f'SiliconCompiler : {chip.design} | {jobname} | {step}{index} | {msg_type}'
    else:
        subject = f'SiliconCompiler : {chip.design} | {jobname} | {msg_type}'

    # Setup email header
    msg['Subject'] = subject

    if "from" in cred:
        msg['From'] = cred["from"]
    else:
        msg['From'] = to[0]
    msg['To'] = ", ".join(to)
    msg['X-Entity-Ref-ID'] = uuid.uuid4().hex  # keep emails from getting grouped

    if cred["max_file_size"] > 0:
        if msg_type == "summary":
            layout_img = report_utils._find_summary_image(chip)
            if layout_img and os.path.isfile(layout_img):
                with open(layout_img, 'rb') as img_file:
                    img_attach = MIMEApplication(img_file.read())
                    img_attach.add_header('Content-Disposition',
                                          'attachment',
                                          filename=os.path.basename(layout_img))
                    msg.attach(img_attach)

            nodes_to_execute = get_executed_nodes(chip, flow)
            nodes, errors, metrics, metrics_unit, metrics_to_show, _ = \
                report_utils._collect_data(chip, flow=flow, flowgraph_nodes=nodes_to_execute)

            text_msg = get_file_template('email/summary.j2').render(
                design=chip.design,
                nodes=nodes,
                errors=errors,
                metrics=metrics,
                metrics_unit=metrics_unit,
                metric_keys=metrics_to_show)
        else:
            # Attach logs
            for log in (f'sc_{step}{index}.log', f'{step}.log'):
                log_file = f'{chip.getworkdir(step=step, index=index)}/{log}'
                if os.path.exists(log_file):
                    with sc_open(log_file) as f:
                        file_content = f.read().splitlines()
                        # Limit to max_file_size
                        file_content = file_content[-cred["max_file_size"]:]
                        log_attach = MIMEApplication("\n".join(file_content))
                        log_name, _ = os.path.splitext(log)
                        # Make attachment a txt file to avoid issues with tools not loading .log
                        log_attach.add_header('Content-Disposition',
                                              'attachment',
                                              filename=f'{log_name}.txt')
                        msg.attach(log_attach)

        records = {}
        for record in chip.getkeys('record'):
            value = None
            if chip.get('record', record, field='pernode').is_never():
                value = chip.get('record', record)
            else:
                value = chip.get('record', record, step=step, index=index)

            if value is not None:
                records[record] = value

        nodes, errors, metrics, metrics_unit, metrics_to_show, _ = \
            report_utils._collect_data(chip, flow=flow, flowgraph_nodes=[(step, index)])

        status = chip.get('record', 'status', step=step, index=index)

        text_msg = get_file_template('email/general.j2').render(
            design=chip.design,
            job=jobname,
            step=step,
            index=index,
            status=status,
            records=records,
            nodes=nodes,
            errors=errors,
            metrics=metrics,
            metrics_unit=metrics_unit,
            metric_keys=metrics_to_show)

    body = MIMEText(text_msg, 'html')
    msg.attach(body)

    if cred['ssl']:
        smtp_use = smtplib.SMTP_SSL
    else:
        smtp_use = smtplib.SMTP

    with smtp_use(cred["server"], cred["port"]) as smtp_server:
        do_send = False
        try:
            smtp_server.login(cred["username"], cred["password"])
            do_send = True
        except smtplib.SMTPAuthenticationError as e:
            chip.logger.error(f'Unable to authenticate to email server: {e}')
        except Exception as e:
            chip.logger.error(f'An error occurred during login to email server: {e}')

        if do_send:
            try:
                smtp_server.sendmail(msg['From'], to, msg.as_string())
            except Exception as e:
                chip.logger.error(f'An error occurred while sending email: {e}')


if __name__ == "__main__":
    from siliconcompiler import Chip
    from siliconcompiler.targets import freepdk45_demo
    chip = Chip('test')
    chip.use(freepdk45_demo)
    chip.set('option', 'scheduler', 'msgevent', 'ALL')
    # chip.set('option', 'scheduler', 'msgcontact', 'fillin')
    send(chip, "BEGIN", "import", "0")



import os
import string
from PIL import Image, ImageFont, ImageDraw

from siliconcompiler import units
from siliconcompiler.report.utils import _find_summary_image, _find_summary_metrics


def _generate_summary_image(chip, output_path):
    '''
    Takes a layout screenshot and generates a design summary image
    featuring a layout thumbnail and several metrics.
    '''

    img_path = _find_summary_image(chip)
    if not img_path:
        return

    # Extract metrics for display
    metrics = {
        'Chip': chip.design,
    }

    pdk = chip.get('option', 'pdk')
    if pdk:
        metrics['Node'] = pdk
    else:
        fpga_partname = chip.get('fpga', 'partname')
        if fpga_partname:
            metrics['FPGA'] = fpga_partname

    def format_area(value, unit):
        prefix = units.get_si_prefix(unit)
        mm_area = units.convert(value, from_unit=prefix, to_unit='mm^2')
        if mm_area < 10:
            return units.format_si(value, 'um') + 'um^2'
        else:
            return units.format_si(mm_area, 'mm') + 'mm^2'

    def format_freq(value, unit):
        value = units.convert(value, from_unit=unit)
        return units.format_si(value, 'Hz') + 'Hz'

    metrics.update(_find_summary_metrics(chip, {
        'Area': ('totalarea', format_area),
        'LUTs': ('luts', None),
        'Fmax': ('fmax', format_freq)}))

    # Generate design

    WIDTH = 1024
    BORDER = 32
    LINE_SPACING = 8
    TEXT_INDENT = 16

    FONT_PATH = os.path.join(chip.scroot, 'data', 'RobotoMono', 'RobotoMono-Regular.ttf')
    FONT_SIZE = 40

    # matches dark gray background color configured in klayout_show.py
    BG_COLOR = (33, 33, 33)

    # near-white
    TEXT_COLOR = (224, 224, 224)

    original_layout = Image.open(img_path)
    orig_width, orig_height = original_layout.size

    aspect_ratio = orig_height / orig_width

    # inset by border left and right
    thumbnail_width = WIDTH - 2 * BORDER
    thumbnail_height = round(thumbnail_width * aspect_ratio)
    layout_thumbnail = original_layout.resize((thumbnail_width, thumbnail_height))

    font = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    # Get max height of any ASCII character in font, so we can consistently space each line
    _, descent = font.getmetrics()
    _, bb_top, _, bb_bottom = font.getmask(string.printable).getbbox()
    line_height = (bb_bottom - bb_top) + descent

    text = []
    x = BORDER + TEXT_INDENT
    y = thumbnail_height + 2 * BORDER
    for metric, value in metrics.items():
        line = f'{metric}: {value}'

        # shorten line till it fits
        cropped_line = line
        while True:
            line_width = font.getmask(cropped_line).getbbox()[2] + TEXT_INDENT
            if x + line_width < (WIDTH - BORDER):
                break
            cropped_line = cropped_line[:-1]

        if cropped_line != line:
            chip.logger.warning(f'Cropped {line} to {cropped_line} to fit in design summary '
                                'image')

        # Stash line to write and coords to write it at
        text.append(((x, y), cropped_line))

        y += line_height + LINE_SPACING

    design_summary = Image.new('RGB', (WIDTH, y + BORDER), color=BG_COLOR)
    design_summary.paste(layout_thumbnail, (BORDER, BORDER))

    draw = ImageDraw.Draw(design_summary)
    for coords, line in text:
        draw.text(coords, line, TEXT_COLOR, font=font)

    design_summary.save(output_path)
    chip.logger.info(f'Generated summary image at {output_path}')


def _open_summary_image(image):
    Image.open(image).show()



from PIL import Image
from siliconcompiler.tools.graphviz import screenshot


def setup(chip):
    '''
    Show a graphviz dot file
    '''

    screenshot.setup(chip)


def run(chip):
    screenshot_ret = screenshot.run(chip)
    if screenshot_ret != 0:
        return screenshot_ret

    Image.open(f"outputs/{chip.top()}.png").show()

    return 0



from siliconcompiler.checklists import oh_tapeout


def get_checklists():
    '''
    Returns a dict of builtin checklists
    '''
    return {
        module.__name__.split(".")[-1]: module for module in (
            oh_tapeout,
        )
    }




'''
GHDL is an open-source analyzer, compiler, simulator and
(experimental) synthesizer for VHDL. It allows you to analyse
and elaborate sources for generating machine code from your design.
Native program execution is the only way for high speed simulation.

Documentation: https://ghdl.readthedocs.io/en/latest

Sources: https://github.com/ghdl/ghdl

Installation: https://github.com/ghdl/ghdl
'''


#####################################################################
# Make Docs
#####################################################################
def make_docs(chip):
    from siliconcompiler.tools.ghdl import convert
    convert.setup(chip)
    return chip


################################
# Version Check
################################
def parse_version(stdout):
    # first line: GHDL 2.0.0-dev (1.0.0.r827.ge49cb7b9) [Dunoon edition]

    # '*-dev' is interpreted by packaging.version as a "developmental release",
    # which has the correct semantics. e.g. Version('2.0.0') > Version('2.0.0-dev')
    return stdout.split()[1]


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("ghdl.json")



#!/usr/bin/env python3

from siliconcompiler import Chip
import cProfile
import pstats
from pstats import SortKey
import graphviz
import argparse
import tempfile
import os
import gprof2dot
from siliconcompiler.targets import freepdk45_demo, skywater130_demo, asap7_demo, asic_demo


def print_stats(pr):
    ps = pstats.Stats(pr).sort_stats(SortKey.CUMULATIVE)
    ps.print_stats(20)


def generate_graph(pr, name):
    with tempfile.TemporaryDirectory() as d:
        stats = os.path.join(d, 'profile.pstats')
        dot_file = os.path.join(d, 'profile.dot')

        pr.dump_stats(stats)

        gprof2dot.main(argv=['-f', 'pstats', stats, '-o', dot_file])

        with open(dot_file) as f:
            dot = graphviz.Source(f.read())

    try:
        dot.render(filename=name, format='png', cleanup=True)
    except graphviz.ExecutableNotFound:
        pass


def run_read_manifest(pr, extra):
    remove_path = False
    if not extra:
        chip = Chip('write')
        chip.use(freepdk45_demo)
        chip.use(skywater130_demo)
        chip.use(asap7_demo)

        fd, path = tempfile.mkstemp(prefix='read_manifest', suffix='.json')
        os.close(fd)

        chip.write_manifest(path)
        remove_path = True
    else:
        path = extra

    chip = Chip('read')

    pr.enable()
    chip.read_manifest(path)
    pr.disable()

    if remove_path:
        os.remove(path)


def run_write_manifest_abspath(pr, extra):
    run_write_manifest(pr, extra, True)


def run_write_manifest(pr, extra, abspath=False):
    chip = Chip('write')
    chip.use(freepdk45_demo)
    chip.use(skywater130_demo)
    chip.use(asap7_demo)
    if extra:
        chip.use(chip._load_module(extra))

    fd, path = tempfile.mkstemp(prefix='write_manifest', suffix='.json')
    os.close(fd)

    pr.enable()
    chip.write_manifest(path, abspath=abspath)
    pr.disable()

    os.remove(path)


def run_check_filepaths(pr, extra):
    chip = Chip('write')
    chip.use(freepdk45_demo)
    chip.use(skywater130_demo)
    chip.use(asap7_demo)
    if extra:
        chip.use(chip._load_module(extra))

    chip.check_filepaths()
    pr.enable()
    for _ in range(10):
        chip.check_filepaths()
    pr.disable()


def run_load_target(pr, extra):
    chip = Chip('write')

    pr.enable()
    if extra:
        chip.use(chip._load_module(extra))
    else:
        chip.use(freepdk45_demo)
        chip.use(skywater130_demo)
        chip.use(asap7_demo)
    pr.disable()


def run_asic_demo(pr, extra):
    chip = Chip('')

    pr.enable()
    chip.use(asic_demo)
    chip.run()
    chip.summary()
    pr.disable()


if __name__ == "__main__":
    tests = {
        'read_manifest': run_read_manifest,
        'write_manifest': run_write_manifest,
        'write_manifest_abspath': run_write_manifest_abspath,
        'load_target': run_load_target,
        'asic_demo': run_asic_demo,
        'check_filepaths': run_check_filepaths,
        'all': None
    }

    parser = argparse.ArgumentParser(
        description='Utility tool to aide in profiling SiliconCompiler')
    parser.add_argument(
        '--print',
        action='store_true',
        help='Print the time stats from the profile')
    parser.add_argument(
        '--profile',
        choices=tests.keys(),
        default='all',
        help='Profile test to run')
    parser.add_argument(
        '--extra',
        metavar='<payload>',
        help='extra payload information for test')

    args = parser.parse_args()

    test_set = args.profile
    if test_set == 'all':
        test_set = [t for t in tests.keys() if t != 'all']
    else:
        test_set = [test_set]

    for test in test_set:
        print(f'Running {test}')
        func = tests[test]
        pr = cProfile.Profile()
        func(pr, args.extra)
        generate_graph(pr, test)
        if args.print:
            print_stats(pr)



from siliconcompiler import NodeStatus
from siliconcompiler.flowgraph import _get_flowgraph_nodes

from pathlib import Path


def get_chip_cwd(chip, manifest):
    build_dir = Path(chip.get('option', 'builddir'))

    manifest_path = Path(manifest)
    for path in manifest_path.parents:
        if build_dir.name == path.name:
            return str(path.parent)

    return None


def make_node_to_step_index_map(chip, metric_dataframe):
    '''
    Returns a map from the name of a node to the associated step, index pair.

    Args:
        metric_dataframe (pandas.DataFrame) : A dataframe full of all metrics and all
            nodes of the selected chip
    '''
    node_to_step_index_map = {}
    for step, index in _get_flowgraph_nodes(chip, chip.get('option', 'flow')):
        node_to_step_index_map[f'{step}{index}'] = (step, index)

    # concatenate step and index
    metric_dataframe.columns = metric_dataframe.columns.map(lambda x: f'{x[0]}{x[1]}')
    return node_to_step_index_map, metric_dataframe


def make_metric_to_metric_unit_map(metric_dataframe):
    '''
    Returns a map from the name of a metric to the associated metric and unit in
    the form f'{x[0]} ({x[1]})'

    Args:
        metric_dataframe (pandas.DataFrame) : A dataframe full of all metrics and all
            nodes of the selected chip.
    '''
    metric_to_metric_unit_map = {}
    for metric, unit in metric_dataframe.index.tolist():
        if unit != '':
            metric_to_metric_unit_map[f'{metric} ({unit})'] = metric
        else:
            metric_to_metric_unit_map[metric] = metric
    # concatenate metric and unit
    metric_dataframe.index = metric_dataframe.index.map(lambda x: f'{x[0]} ({x[1]})'
                                                        if x[1] else x[0])
    return metric_to_metric_unit_map, metric_dataframe


def is_running(chip):
    for step, index in _get_flowgraph_nodes(chip, chip.get('option', 'flow')):
        state = chip.get('record', 'status', step=step, index=index)
        if not NodeStatus.is_done(state):
            return True
    return False


def generate_metric_dataframe(chip):
    from siliconcompiler.report import report

    metric_dataframe = report.make_metric_dataframe(chip)
    node_to_step_index_map, metric_dataframe = \
        make_node_to_step_index_map(chip, metric_dataframe)
    metric_to_metric_unit_map, metric_dataframe = \
        make_metric_to_metric_unit_map(metric_dataframe)

    return metric_dataframe, node_to_step_index_map, metric_to_metric_unit_map



import shutil
import os

from siliconcompiler.tools.openroad import make_docs as or_make_docs
from siliconcompiler.tools.openroad._apr import setup as tool_setup
from siliconcompiler.tools.openroad._apr import set_reports
from siliconcompiler.tools.openroad._apr import build_pex_corners, define_ord_files
from siliconcompiler.tools._common import find_incoming_ext, input_provides, get_tool_task
from siliconcompiler.tools.openroad._apr import \
    define_ord_params, define_sta_params, define_sdc_params


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    or_make_docs(chip)
    chip.set('tool', 'openroad', 'task', 'show', 'var', 'show_filepath', '<path>')


def setup(chip):
    '''
    Show a design in openroad
    '''
    generic_show_setup(chip, False)


def generic_show_setup(chip, exit):
    # Generic tool setup.
    tool_setup(chip, exit=exit)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    design = chip.top()

    chip.set('tool', tool, 'task', task, 'script', 'sc_show.tcl',
             step=step, index=index)

    # Add GUI option
    chip.add('tool', tool, 'task', task, 'option', '-gui',
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'show_exit', exit,
             step=step, index=index, clobber=False)
    if chip.valid('tool', tool, 'task', task, 'var', 'show_filepath'):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['tool', tool, 'task', task, 'var', 'show_filepath']),
                 step=step, index=index)
    else:
        incoming_ext = find_incoming_ext(chip, ('odb', 'def'), 'def')
        chip.set('tool', tool, 'task', task, 'var', 'show_filetype', incoming_ext,
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'input', f'{design}.{incoming_ext}',
                 step=step, index=index)
        if f'{design}.sdc' in input_provides(chip, step, index):
            chip.add('tool', tool, 'task', task, 'input', f'{design}.sdc',
                     step=step, index=index)

    # set default values for task
    define_ord_params(chip)
    define_sta_params(chip)
    define_sdc_params(chip)

    set_reports(chip, [])


def pre_process(chip):
    copy_show_files(chip)
    define_ord_files(chip)
    build_pex_corners(chip)


def copy_show_files(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if chip.valid('tool', tool, 'task', task, 'var', 'show_filepath'):
        show_file = chip.get('tool', tool, 'task', task, 'var', 'show_filepath',
                             step=step, index=index)[0]
        show_type = chip.get('tool', tool, 'task', task, 'var', 'show_filetype',
                             step=step, index=index)[0]
        show_job = chip.get('tool', tool, 'task', task, 'var', 'show_job',
                            step=step, index=index)
        show_step = chip.get('tool', tool, 'task', task, 'var', 'show_step',
                             step=step, index=index)
        show_index = chip.get('tool', tool, 'task', task, 'var', 'show_index',
                              step=step, index=index)

        # copy source in to keep sc_apr.tcl simple
        dst_file = f"inputs/{chip.top()}.{show_type}"
        shutil.copy2(show_file, dst_file)
        if show_job and show_step and show_index:
            sdc_file = chip.find_result('sdc', show_step[0],
                                        jobname=show_job[0],
                                        index=show_index[0])
            if sdc_file and os.path.exists(sdc_file):
                shutil.copy2(sdc_file, f"inputs/{chip.top()}.sdc")



import pya
import sys
import os


def read_layout(stream_file):
    print(f"[INFO] Reading '{stream_file}'")
    layout = pya.Layout()
    layout.read(stream_file)

    return layout


def __with_timestamps(schema):
    sc_step = schema.get('arg', 'step')
    sc_index = schema.get('arg', 'index')

    return schema.get('tool', 'klayout', 'task', 'operations', 'var', 'timestamps',
                      step=sc_step, index=sc_index) == ['true']


def __get_keypath_step_index(schema, *keypath):
    ret = {
        'step': schema.get('arg', 'step'),
        'index': schema.get('arg', 'index')
    }
    pernode = schema.get(*keypath, field='pernode')
    if pernode.is_never():
        ret['step'] = None
        ret['index'] = None

    return ret


def __do_cell_swap(parent, old_cell_idx, new_cell, checked):
    if (parent.cell_index() in checked):
        return 0

    checked.append(parent.cell_index())
    replacements = 0
    for inst in parent.each_inst():
        if (inst.cell_index == old_cell_idx):
            inst.cell = new_cell
            replacements += 1
        else:
            replacements += __do_cell_swap(inst.cell, old_cell_idx, new_cell, checked)
    return replacements


def swap_cells(base_layout, oldcell, newcell):
    top_cell = base_layout.top_cell()
    old_cell = base_layout.cell(oldcell)
    new_cell = base_layout.cell(newcell)

    if (old_cell is None):
        return base_layout
    if (new_cell is None):
        return base_layout

    checked = []
    replacements = __do_cell_swap(top_cell, old_cell.cell_index(), new_cell, checked)
    print(f"[INFO] Swapping '{old_cell.name}' to '{new_cell.name}' in "
          f"'{top_cell.name}': {replacements} occurrences")
    base_layout.delete_cell(old_cell.cell_index())

    return base_layout


def add_outline(base_layout, layer):
    top_cell = base_layout.top_cell()
    bbox = top_cell.bbox()

    layer_info = base_layout.get_info(layer)
    print(f"[INFO] Adding outline to '{top_cell.name}' on layer '{layer_info.to_s()}'")

    shapes = top_cell.shapes(layer)
    shapes.insert(pya.Box(bbox))

    return base_layout


def add_layout(base_layout, layout):
    top_cell = base_layout.top_cell()

    other_layout_top = layout.top_cell()

    print(f"[INFO] Adding layout from '{other_layout_top.name}' to '{top_cell.name}'")
    new_cell = base_layout.create_cell(other_layout_top.name)
    new_cell.copy_tree(other_layout_top)

    cell_inst = pya.CellInstArray(new_cell.cell_index(), pya.Trans())
    top_cell.insert(cell_inst)

    return base_layout


def add_layout_to_top(base_layout, new_top_cell_name):
    top_cell = base_layout.top_cell()

    print(f"[INFO] Adding layout from '{top_cell.name}' to new top cell '{new_top_cell_name}'")
    new_cell = base_layout.create_cell(new_top_cell_name)

    cell_inst = pya.CellInstArray(top_cell.cell_index(), pya.Trans())
    new_cell.insert(cell_inst)

    return base_layout


def merge_layouts(layout1, layout2):
    cell1 = layout1.top_cell()
    cell2 = layout2.top_cell()

    print(f"[INFO] Merging cells '{cell1.name}' and '{cell2.name}' into '{cell1.name}'")

    cell1.copy_tree(cell2)

    return layout1


def rotate_layout(base_layout):
    top_cell = base_layout.top_cell()
    bbox = top_cell.bbox()

    print(f"[INFO] Rotating layout '{top_cell.name}' 90 degrees")

    transform = pya.Trans.R270
    transform = pya.Trans(transform, pya.Vector(0, bbox.p2.x))

    top_cell.transform(transform)

    return base_layout


def rename_top(base_layout, new_name):
    top_cell = base_layout.top_cell()
    print(f"[INFO] Renaming '{top_cell.name}' to '{new_name}' layout: '{top_cell.name}'")
    top_cell.name = new_name
    return base_layout


def rename_cell(base_layout, old_name, new_name):
    cell = base_layout.cell(old_name)
    if not cell:
        print(f"[WARNING] Unable to find '{old_name}' to rename")
    print(f"[INFO] Renaming '{cell.name}' to '{new_name}' layout: '{base_layout.top_cell().name}'")
    cell.name = new_name
    return base_layout


def write_stream(layout, outfile, timestamps):
    from klayout_utils import get_write_options

    print(f"[INFO] Writing layout: '{outfile}'")

    layout.write(outfile, get_write_options(outfile, timestamps))


def make_property_text(layout, property_layer, property_name, destination_layer):
    property_layer_info = layout.get_info(property_layer)
    destination_layer_info = layout.get_info(destination_layer)
    print(f"[INFO] Generating properties from {property_layer_info.to_s()} "
          f"/ {property_name} on {destination_layer_info.to_s()}")

    top_cell = layout.top_cell()
    # Generate list of text objects
    source_shapes_itr = top_cell.begin_shapes_rec(property_layer)
    dest_shapes = []
    while (not source_shapes_itr.at_end()):
        shape = source_shapes_itr.shape()
        shape_prop = shape.property(property_name)
        if (shape_prop is not None and (shape.is_box() or shape.is_polygon())):
            shape_center = shape.bbox().center()
            dest_shapes.append(pya.Text(shape_prop, shape_center.x, shape_center.y))
        source_shapes_itr.next()

    # Insert objects
    dest_shapes_layer = top_cell.shapes(destination_layer)
    for shape in dest_shapes:
        dest_shapes_layer.insert(shape)

    print(f"[INFO] Generated {len(dest_shapes)} text shapes.")

    return layout


def delete_layers(layout, layers):
    for cell in layout.each_cell():
        print(f'[INFO] Deleting layers from {cell.name}')
        for layer in layers:
            layer_info = layout.get_info(layer)
            print(f"[INFO] Deleting layer {layer_info.to_s()}")

            cell.shapes(layer).clear()

    return layout


def merge_shapes(layout, layers):
    if layers == ['all']:
        layers = layout.layer_indexes()

    for cell in layout.each_cell():
        print(f"[INFO] Merging shapes in {cell.name}")
        for layer in layers:
            layer_info = layout.get_info(layer)
            print(f"[INFO] Merging shapes on layer {layer_info.to_s()}")

            shape_proc = pya.ShapeProcessor()
            output_shapes = pya.Shapes()

            cell_layout = cell.layout()
            print("  Shape count (old):", cell.shapes(layer).size())
            shape_proc.boolean(cell_layout,
                               cell,
                               layer,
                               cell_layout,
                               cell,
                               layer,
                               output_shapes,
                               pya.EdgeProcessor.ModeOr,
                               True,
                               True,
                               True)
            print("  Shape count (new):", output_shapes.size())
            cell.shapes(layer).clear()
            cell.shapes(layer).insert(output_shapes)


def flatten(layout):
    top_cell = layout.top_cell()

    print(f"[INFO] Flattening: {top_cell.name}")
    top_cell.flatten(True)

    return layout


def parse_operations(schema, base_layout, steps):
    for step in steps:
        step = step.split(":")
        step_name = step[0]
        step_args = ":".join(step[1:])
        args_key = step_args.split(',')

        if (step_name == "merge" or step_name == "add"):
            files = []
            if len(args_key) > 1:
                if 'file' not in schema.get(*args_key, field='type'):
                    raise ValueError(f'{step_name} requires {args_key} be a file type')
                files = schema.get(*args_key, **__get_keypath_step_index(schema, *args_key))
            else:
                files = [f'inputs/{step_args}']
            for op_file in files:
                if step_name == "add":
                    base_layout = add_layout(base_layout, read_layout(op_file))
                else:
                    base_layout = merge_layouts(base_layout, read_layout(op_file))
        elif (step_name == "rotate"):
            base_layout = rotate_layout(base_layout)
        elif (step_name == "outline"):
            outline_layer = [int(layer) for layer in schema.get(
                *args_key, **__get_keypath_step_index(schema, *args_key))]
            if len(outline_layer) != 2:
                raise ValueError('outline layer requires two entries for layer and purpose, '
                                 f'received: {len(outline_layer)}')
            base_layout = add_outline(base_layout,
                                      base_layout.layer(outline_layer[0], outline_layer[1]))
        elif (step_name == "convert_property"):
            options = schema.get(*args_key, **__get_keypath_step_index(schema, *args_key))
            if len(options) != 3 and len(options) != 5:
                raise ValueError(f'{step_name} requires 3 or 5 arguments in {args_key}')
            prop_layer = [int(layer) for layer in options[0:2]]
            prop_number = options[2]
            if prop_number.isnumeric():
                prop_number = int(prop_number)
            if (len(options) == 5):
                dest_layer = [int(layer) for layer in options[3:]]
            else:
                dest_layer = prop_layer
            base_layout = make_property_text(base_layout,
                                             base_layout.layer(prop_layer[0], prop_layer[1]),
                                             prop_number,
                                             base_layout.layer(dest_layer[0], dest_layer[1]))
        elif (step_name == "rename"):
            new_name = schema.get(*args_key, **__get_keypath_step_index(schema, *args_key))[0]
            base_layout = rename_top(base_layout, new_name)
        elif (step_name == "rename_cell"):
            new_name = schema.get(*args_key, **__get_keypath_step_index(schema, *args_key))[0]
            for renameset in schema.get(*args_key, **__get_keypath_step_index(schema, *args_key)):
                oldcell, newcell = renameset.split("=")
                base_layout = rename_cell(base_layout, oldcell, newcell)
        elif (step_name == "swap"):
            for swapset in schema.get(*args_key, **__get_keypath_step_index(schema, *args_key)):
                oldcell, newcell = swapset.split("=")
                base_layout = swap_cells(base_layout, oldcell, newcell)
        elif (step_name == "add_top"):
            new_name = schema.get(*args_key, **__get_keypath_step_index(schema, *args_key))[0]
            base_layout = add_layout_to_top(base_layout, new_name)
        elif (step_name == "write"):
            write_stream(base_layout, f'outputs/{step_args}', __with_timestamps(schema))
        elif (step_name == "flatten"):
            base_layout = flatten(base_layout)
        elif (step_name == "delete_layers"):
            layers = []
            for layer in schema.get(*args_key, **__get_keypath_step_index(schema, *args_key)):
                layer_num = None
                layer_purpose = None
                if '/' in layer:
                    layer_num, layer_purpose = layer.split('/')
                elif ' ' in layer:
                    layer_num, layer_purpose = layer.split(' ')
                elif ':' in layer:
                    layer_num, layer_purpose = layer.split(':')
                else:
                    raise ValueError(f'Unable to determine layer purpose pair for {layer}')
                layers.append(base_layout.layer(int(layer_num), int(layer_purpose)))
            base_layout = delete_layers(base_layout, layers)
        elif (step_name == "merge_shapes"):
            layers = schema.get(*args_key, **__get_keypath_step_index(schema, *args_key))
            base_layout = merge_shapes(base_layout, layers)
        else:
            raise ValueError(f"Unknown step: {step_name}")


if __name__ == "__main__":
    # SC_ROOT provided by CLI
    sys.path.append(SC_KLAYOUT_ROOT)  # noqa: F821
    sys.path.append(SC_TOOLS_ROOT)  # noqa: F821

    from klayout_utils import (
        technology,
        get_streams,
        get_schema,
        generate_metrics
    )

    schema = get_schema(manifest='sc_manifest.json')

    # Extract info from manifest
    sc_step = schema.get('arg', 'step')
    sc_index = schema.get('arg', 'index')
    sc_tool = 'klayout'
    sc_task = 'operations'

    sc_ext = get_streams(schema)[0]
    design = schema.get('option', 'entrypoint')
    if not design:
        design = schema.get('design')

    in_gds = os.path.join('inputs', f'{design}.{sc_ext}.gz')
    if not os.path.exists(in_gds):
        in_gds = os.path.join('inputs', f'{design}.{sc_ext}')
    if not os.path.exists(in_gds):
        in_gds = schema.get('input', 'layout', sc_ext)[0]
    out_gds = os.path.join('outputs', f'{design}.{sc_ext}')

    tech = technology(design, schema)
    base_layout = read_layout(in_gds)
    base_layout.technology_name = tech.name

    sc_klayout_ops = schema.get('tool', sc_tool, 'task', sc_task, 'var', 'operations',
                                step=sc_step, index=sc_index)
    parse_operations(schema, base_layout, sc_klayout_ops)

    write_stream(base_layout, out_gds, __with_timestamps(schema))

    generate_metrics()



import siliconcompiler

from siliconcompiler.tools.icarus import compile as icarus_compile
from siliconcompiler.tools.verilator import compile as verilator_compile
from siliconcompiler.tools.execute import exec_input
from siliconcompiler.tools.xyce import simulate as xyce_simulate
from siliconcompiler.tools.xdm import convert as xdm_convert


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    chip.set('input', 'rtl', 'netlist', 'test')
    return setup(np=5)


#############################################################################
# Flowgraph Setup
#############################################################################
def setup(flowname='dvflow',
          tool='icarus',
          np=1):
    '''
    A configurable constrained random stimulus DV flow.

    The verification pipeline includes the following steps:

    * **compile**: RTL sources are compiled into object form (once)
    * **sim**: Compiled RTL is exercised using generated test

    The dvflow can be parametrized using a single 'np' parameter.
    Setting 'np' > 1 results in multiple independent verification
    pipelines to be launched.

    Supported tools are:

    * icarus
    * verilator
    * xyce
    * xdm-xyce

    This flow is a WIP
    '''

    flow = siliconcompiler.Flow(flowname)

    tasks = {}
    flow_np = {}

    if tool == 'icarus':
        tasks['compile'] = icarus_compile
        tasks['sim'] = exec_input

        flowpipe = [
            'compile',
            'sim'
        ]
        flow_np = {
            'compile': 1,
            'sim': np
        }
    elif tool == 'verilator':
        tasks['compile'] = verilator_compile
        tasks['sim'] = exec_input

        flowpipe = [
            'compile',
            'sim'
        ]
        flow_np = {
            'compile': 1,
            'sim': np
        }
    elif tool == 'xyce':
        tasks['sim'] = xyce_simulate

        flowpipe = [
            'sim'
        ]
        flow_np = {
            'sim': np
        }
    elif tool == 'xdm-xyce':
        tasks['compile'] = xdm_convert
        tasks['sim'] = xyce_simulate

        flowpipe = [
            'compile',
            'sim'
        ]
        flow_np = {
            'compile': 1,
            'sim': np
        }
    else:
        raise ValueError(f'{tool} is not a supported tool for {flowname}')

    prevstep = None
    # Flow setup
    for step in flowpipe:
        task = tasks[step]

        parallel = flow_np[step]

        for n in range(parallel):
            flow.node(flowname, step, task, index=n)

            if prevstep:
                flow.edge(flowname, prevstep, step, tail_index=0, head_index=n)

        prevstep = step

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



import os
from urllib.parse import urlparse
import importlib
import re
from siliconcompiler import SiliconCompilerError
from siliconcompiler.utils import default_cache_dir, _resolve_env_vars
import json
from importlib.metadata import distributions, distribution
import functools
import time
from pathlib import Path

from github import Github
import github.Auth

from siliconcompiler.utils import get_plugins


def get_cache_path(chip):
    cache_path = chip.get('option', 'cachedir')
    if cache_path:
        cache_path = chip.find_files('option', 'cachedir', missing_ok=True)
        if not cache_path:
            cache_path = os.path.join(chip.cwd, chip.get('option', 'cachedir'))
    if not cache_path:
        cache_path = default_cache_dir()

    return cache_path


def get_download_cache_path(chip, package, ref):
    cache_path = get_cache_path(chip)
    if not os.path.exists(cache_path):
        os.makedirs(cache_path, exist_ok=True)

    if ref is None:
        raise SiliconCompilerError(f'Reference is required for cached data: {package}', chip=chip)

    return \
        os.path.join(cache_path, f'{package}-{ref}'), \
        os.path.join(cache_path, f'{package}-{ref}.lock')


def _file_path_resolver(chip, package, path, ref, url):
    return os.path.abspath(path.replace('file://', ''))


def _python_path_resolver(chip, package, path, ref, url):
    return path_from_python(chip, url.netloc)


def _get_path_resolver(path):
    url = urlparse(path)

    for resolver in get_plugins("path_resolver"):
        func = resolver(url)
        if func:
            return func, url

    if url.scheme == "file":
        return _file_path_resolver, url

    if url.scheme == "python":
        return _python_path_resolver, url

    raise ValueError(f"{path} is not supported")


def _path(chip, package):
    # Initially try retrieving data source from schema
    data = {}
    data['path'] = chip.get('package', 'source', package, 'path')
    data['ref'] = chip.get('package', 'source', package, 'ref')
    if not data['path']:
        raise SiliconCompilerError(
            f'Could not find package source for {package} in schema. '
            'You can use register_source() to add it.', chip=chip)

    data['path'] = _resolve_env_vars(chip, data['path'], None, None)

    if os.path.exists(data['path']):
        # Path is already a path
        return os.path.abspath(data['path'])

    path_resolver, url = _get_path_resolver(data['path'])

    return path_resolver(chip, package, data['path'], data['ref'], url)


def path(chip, package):
    """
    Compute data source data path
    Additionally cache data source data if possible
    Parameters:
        package (str): Name of the data source
    Returns:
        path: Location of data source on the local system
    """

    if package not in chip._packages:
        changed = False
        data_path = _path(chip, package)

        if isinstance(data_path, tuple) and len(data_path) == 2:
            data_path, changed = data_path

        if os.path.exists(data_path):
            if package not in chip._packages and changed:
                chip.logger.info(f'Saved {package} data to {data_path}')
            else:
                chip.logger.info(f'Found {package} data at {data_path}')

            chip._packages[package] = data_path
        else:
            raise SiliconCompilerError(f'Unable to locate {package} data in {data_path}',
                                       chip=chip)

    return chip._packages[package]


def __get_filebased_lock(data_lock):
    base, _ = os.path.splitext(os.fsdecode(data_lock.path))
    return Path(f'{base}.sc_lock')


def aquire_data_lock(data_path, data_lock):
    # Wait a maximum of 10 minutes for other processes to finish
    max_seconds = 10 * 60
    try:
        if data_lock.acquire(timeout=max_seconds):
            return
    except RuntimeError:
        # Fall back to file based locking method
        lock_file = __get_filebased_lock(data_lock)
        while (lock_file.exists()):
            if max_seconds == 0:
                raise SiliconCompilerError(f'Failed to access {data_path}.'
                                           f'Lock {lock_file} still exists.')
            time.sleep(1)
            max_seconds -= 1

        lock_file.touch()

        return

    raise SiliconCompilerError(f'Failed to access {data_path}. '
                               f'{data_lock.path} is still locked, if this is a mistake, '
                               'please delete it.')


def release_data_lock(data_lock):
    # Check if file based locking method was used
    lock_file = __get_filebased_lock(data_lock)
    if lock_file.exists():
        lock_file.unlink(missing_ok=True)
        return

    data_lock.release()


def path_from_python(chip, python_package, append_path=None):
    try:
        module = importlib.import_module(python_package)
    except:  # noqa E722
        raise SiliconCompilerError(f'Failed to import {python_package}.', chip=chip)

    python_path = os.path.dirname(module.__file__)
    if append_path:
        if isinstance(append_path, str):
            append_path = [append_path]
        python_path = os.path.join(python_path, *append_path)

    python_path = os.path.abspath(python_path)

    return python_path


def is_python_module_editable(module_name):
    dist_map = __get_python_module_mapping()
    dist = dist_map[module_name][0]

    is_editable = False
    for f in distribution(dist).files:
        if f.name == 'direct_url.json':
            info = None
            with open(f.locate(), 'r') as f:
                info = json.load(f)

            if "dir_info" in info:
                is_editable = info["dir_info"].get("editable", False)

    return is_editable


def register_python_data_source(chip,
                                package_name,
                                python_module,
                                alternative_path,
                                alternative_ref=None,
                                python_module_path_append=None):
    '''
    Helper function to register a python module as data source with an alternative in case
    the module is not installed in an editable state
    '''
    # check if installed in an editable state
    if is_python_module_editable(python_module):
        if python_module_path_append:
            path = path_from_python(chip, python_module, append_path=python_module_path_append)
        else:
            path = f"python://{python_module}"
        ref = None
    else:
        path = alternative_path
        ref = alternative_ref

    chip.register_source(name=package_name,
                         path=path,
                         ref=ref)


def register_private_github_data_source(chip,
                                        package_name,
                                        repository,
                                        release,
                                        artifact):
    gh = Github(auth=github.Auth.Token(__get_github_auth_token(package_name)))
    repo = gh.get_repo(repository)

    if not release:
        release = repo.get_latest_release().tag_name

    url = None
    for repo_release in repo.get_releases():
        if repo_release.tag_name == release:
            for asset in repo_release.assets:
                if asset.name == artifact:
                    url = asset.url

    if not url:
        raise ValueError(f'Unable to find release asset: {repository}/{release}/{artifact}')

    chip.register_source(
        package_name,
        path=url,
        ref=release)


def __get_github_auth_token(package_name):
    token_name = package_name.upper()
    for tok in ('#', '$', '&', '-', '=', '!', '/'):
        token_name = token_name.replace(tok, '')

    search_env = (
        f'GITHUB_{token_name}_TOKEN',
        'GITHUB_TOKEN',
        'GIT_TOKEN'
    )

    token = None
    for env in search_env:
        token = os.environ.get(env, None)

        if token:
            break

    if not token:
        raise ValueError('Unable to determine authorization token for GitHub, '
                         f'please set one of the following environmental variables: {search_env}')

    return token


@functools.lru_cache(maxsize=1)
def __get_python_module_mapping():
    mapping = {}

    for dist in distributions():
        dist_name = None
        if hasattr(dist, 'name'):
            dist_name = dist.name
        else:
            metadata = dist.read_text('METADATA')
            if metadata:
                find_name = re.compile(r'Name: (.*)')
                for data in metadata.splitlines():
                    group = find_name.findall(data)
                    if group:
                        dist_name = group[0]
                        break

        if not dist_name:
            continue

        provides = dist.read_text('top_level.txt')
        if provides:
            for module in dist.read_text('top_level.txt').split():
                mapping.setdefault(module, []).append(dist_name)

    return mapping



'''
Builtin tools for SiliconCompiler
'''
from siliconcompiler.tools._common import input_provides, get_tool_task


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    return chip


def set_io_files(chip, inputs=True, outputs=True):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if inputs:
        chip.set('tool', tool, 'task', task, 'input',
                 list(input_provides(chip, step, index).keys()),
                 step=step, index=index)
    if outputs:
        chip.set('tool', tool, 'task', task, 'output',
                 list(input_provides(chip, step, index).keys()),
                 step=step, index=index)



'''
GTKWave is a fully featured GTK+ based wave viewer for Unix, Win32, and
Mac OSX which reads LXT, LXT2, VZT, FST, and GHW files as well as standard
Verilog VCD/EVCD files and allows their viewing.

Documentation: https://gtkwave.github.io/gtkwave/

Sources: https://github.com/gtkwave/gtkwave

Installation: https://github.com/gtkwave/gtkwave
'''

from siliconcompiler.tools._common import \
    get_tool_task


def setup(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', 'gtkwave')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=v3.3.116', clobber=False)
    chip.set('tool', tool, 'format', 'tcl', clobber=False)


################################
# Version Check
################################
def parse_version(stdout):
    # First line: GTKWave Analyzer v3.3.116 (w)1999-2023 BSI
    return stdout.split()[2]


def normalize_version(version):
    if version[0] == 'v':
        return version[1:]
    return version



'''
VPR (Versatile Place and Route) is an open source CAD
tool designed for the exploration of new FPGA architectures and
CAD algorithms, at the packing, placement and routing phases of
the CAD flow. VPR takes, as input, a description of an FPGA
architecture along with a technology-mapped user circuit. It
then performs packing, placement, and routing to map the
circuit onto the FPGA. The output of VPR includes the FPGA
configuration needed to implement the circuit and statistics about
the final mapped design (eg. critical path delay, area, etc).

Documentation: https://docs.verilogtorouting.org/en/latest

Sources: https://github.com/verilog-to-routing/vtr-verilog-to-routing

Installation: https://github.com/verilog-to-routing/vtr-verilog-to-routing
'''

import glob
import os
import shutil
import json
import re
from siliconcompiler import sc_open, SiliconCompilerError
from siliconcompiler.tools._common import get_tool_task, record_metric


__block_file = "reports/block_usage.json"


######################################################################
# Make Docs
######################################################################
def make_docs(chip):
    from siliconcompiler.targets import fpgaflow_demo
    chip.set('fpga', 'partname', 'example_arch_X005Y005')
    chip.use(fpgaflow_demo)
    setup_tool(chip)
    return chip


def setup_tool(chip, clobber=True):

    chip.set('tool', 'vpr', 'exe', 'vpr', clobber=clobber)
    chip.set('tool', 'vpr', 'vswitch', '--version')
    chip.set('tool', 'vpr', 'version', '>=9.0.0', clobber=clobber)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', "^Warning",
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', "^Error",
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'enable_images', 'true',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'enable_images',
             'true/false generate images of the design at the end of the task', field='help')

    part_name = chip.get('fpga', 'partname')
    for resource in ('vpr_registers', 'vpr_brams', 'vpr_dsps'):
        if not chip.valid('fpga', part_name, 'var', resource):
            continue
        if not chip.get('fpga', part_name, 'var', resource):
            continue
        chip.add('tool', tool, 'task', task, 'require', f'fpga,{part_name},var,{resource}',
                 step=step, index=index)

    add_tool_requirements(chip)


def add_tool_requirements(chip):

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    part_name = chip.get('fpga', 'partname')

    chip.add('tool', tool, 'task', task, 'require', f'fpga,{part_name},var,vpr_device_code',
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'timing_paths',
             'number of timing paths to report', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'timing_paths', '20',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'timing_report_type',
             'type of timing report', field='help')
    chip.set('tool', tool, 'task', task, 'var', 'timing_report_type', 'aggregated',
             step=step, index=index, clobber=False)


def runtime_options(chip):

    part_name = chip.get('fpga', 'partname')
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    options = []

    device_code = chip.get('fpga', part_name, 'var', 'vpr_device_code')

    options.append(f"--device {device_code[0]}")

    # Medium-term solution:  VPR performs hash digest checks that
    # fail if file paths are changed between steps.  We wish to
    # disable the digest checks to work around this
    options.append("--verify_file_digests off")

    options.append(f"--write_block_usage {__block_file}")
    options.append("--outfile_prefix outputs/")

    if chip.valid('fpga', part_name, 'file', 'archfile') and \
       chip.get('fpga', part_name, 'file', 'archfile'):

        archs = chip.find_files('fpga', part_name, 'file', 'archfile',
                                step=None, index=None)

    else:
        archs = []

    if (len(archs) == 1):
        options.append(archs[0])
    elif (len(archs) == 0):
        raise SiliconCompilerError(
            "VPR requires an architecture file as one of its command line arguments",
            chip=chip)
    else:
        raise SiliconCompilerError(
            "Only one architecture XML file can be passed to VPR", chip=chip)

    threads = chip.get('tool', tool, 'task', task, 'threads', step=step, index=index)
    options.append(f"--num_workers {threads}")

    # For most architectures, constant nets need to be routed
    # like regular nets to be functionally correct (however inefficient
    # that might be...); these two options help control that
    options.append('--constant_net_method route')
    options.append('--const_gen_inference none')

    # If we allow VPR to sweep dangling primary I/Os and logic blocks
    # it can interfere with circuit debugging; so disable that
    options.append('--sweep_dangling_primary_ios off')
    # If you don't sweep dangling primary I/Os, but sweeping nets
    # VPR can crash:
    options.append('--sweep_dangling_nets off')
    # If you don't sweep dangling nets then the timing engine requires
    # you to set an option allowing dangling nodes
    options.append('--allow_dangling_combinational_nodes on')
    options.append('--sweep_constant_primary_outputs off')
    options.append('--sweep_dangling_blocks off')

    # Explicitly specify the clock modeling type in the part driver
    # to avoid ambiguity and future-proof against new VPR clock models
    clock_model = chip.get('fpga', part_name, 'var', 'vpr_clock_model')
    if not clock_model:
        raise SiliconCompilerError(f'no clock model defined for {part_name}', chip=chip)
    else:
        selected_clock_model = clock_model[0]
        # When dedicated networks are used, tell VPR to use the two-stage router,
        # otherwise not.
        if (selected_clock_model == 'ideal'):
            options.append(f'--clock_modeling {selected_clock_model}')
        elif (selected_clock_model == 'route'):
            options.append(f'--clock_modeling {selected_clock_model}')
        elif (selected_clock_model == 'dedicated_network'):
            options.append(f'--clock_modeling {selected_clock_model}')
            options.append('--two_stage_clock_routing')
        else:
            raise SiliconCompilerError(
                'vpr_clock model must be set to ideal, route, or dedicated_clock_network',
                chip=chip)

    sdc_file = None
    if chip.valid('input', 'constraint', 'sdc'):
        sdc_file = find_single_file(chip, 'input', 'constraint', 'sdc',
                                    step=step, index=index,
                                    file_not_found_msg="SDC file not found")

    if sdc_file:
        sdc_arg = f"--sdc_file {sdc_file}"
        options.append(sdc_arg)

        report_type = chip.get('tool', tool, 'task', task, 'var', 'timing_report_type',
                               step=step, index=index)[0]
        options.append(f'--timing_report_detail {report_type}')
        report_paths = chip.get('tool', tool, 'task', task, 'var', 'timing_paths',
                                step=step, index=index)[0]
        options.append(f'--timing_report_npaths {report_paths}')
    else:
        options.append("--timing_analysis off")

    # Per the scheme implemented in the placement pre-process step,
    # if a constraints file exists it will always be in the auto_constraints()
    # location:
    if (os.path.isfile(auto_constraints())):
        pin_constraint_arg = f"--read_vpr_constraints {auto_constraints()}"
        options.append(pin_constraint_arg)

    # Routing graph XML:
    rr_graph = find_single_file(chip, 'fpga', part_name, 'file', 'graphfile',
                                step=None, index=None,
                                file_not_found_msg="VPR RR Graph not found")
    if (rr_graph is None):
        chip.logger.info("No VPR RR graph file specified")
        chip.logger.info("Routing architecture will come from architecture XML file")
    else:
        options.append("--read_rr_graph " + rr_graph)

    # ***NOTE: For real FPGA chips you need to specify the routing channel
    #          width explicitly.  VPR requires an explicit routing channel
    #          with when --read_rr_graph is used (typically the case for
    #          real chips).  Otherwise VPR performs a binary search for
    #          the minimum routing channel width that the circuit fits in.
    # Given the above, it may be appropriate to couple these variables somehow,
    # but --route_chan_width CAN be used by itself.
    num_routing_channels = chip.get('fpga', part_name, 'var', 'channelwidth')

    if (len(num_routing_channels) == 0):
        raise SiliconCompilerError("Number of routing channels not specified", chip=chip)
    elif (len(num_routing_channels) == 1):
        options.append("--route_chan_width " + num_routing_channels[0])
    elif (len(num_routing_channels) > 1):
        raise SiliconCompilerError(
            "Only one routing channel width argument can be passed to VPR", chip=chip)

    return options


############################
# Get common graphics files
############################

def get_common_graphics(chip):

    design = chip.top()

    graphics_commands = []

    # set_draw_block_text 1 displays the label for various blocks in the design
    # set_draw_block_outlines 1 displays the outline/boundary for various blocks in the design
    # save_graphics saves the block diagram as a png/svg/pdf
    # Refer: https://docs.verilogtorouting.org/en/latest/vpr/command_line_usage/#graphics-options
    graphics_commands.append("set_draw_block_text 1; " +
                             "set_draw_block_outlines 1; " +
                             f"save_graphics reports/{design}_place.png;")

    return graphics_commands


################################
# Wrapper around find files to
# help with error checking that
# only a single file is found
################################

def find_single_file(chip, *keypath, step=None, index=None, file_not_found_msg="File not found"):

    if chip.valid(*keypath) and chip.get(*keypath, step=step, index=index):
        file_list = chip.find_files(*keypath, step=step, index=index)
    else:
        file_list = []

    if (len(file_list) == 1):
        return file_list[0]
    elif (len(file_list) == 0):
        chip.logger.info(file_not_found_msg)
        return None
    else:
        raise SiliconCompilerError("Only one file of this type can be passed to VPR", chip=chip)


################################
# Version Check
################################


def parse_version(stdout):

    # Example output of vpr --version:
    # Note that blank comment lines in this example
    # represent newlines printed by vpr --version

    # VPR FPGA Placement and Routing.
    # Version: 8.1.0-dev+c4156f225
    # Revision: v8.0.0-7887-gc4156f225
    # Compiled: 2023-06-14T17:32:05
    # Compiler: GNU 9.4.0 on Linux-5.14.0-1059-oem x86_64
    # Build Info: release IPO VTR_ASSERT_LEVEL=2
    #
    # University of Toronto
    # verilogtorouting.org
    # vtr-users@googlegroups.com
    # This is free open source code under MIT license.
    #
    #
    return stdout.split()[6]


def normalize_version(version):
    if '-' in version:
        return version.split('-')[0]
    else:
        return version


def auto_constraints():
    return 'inputs/sc_constraints.xml'


def vpr_post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    for report in glob.glob("*.rpt"):
        shutil.move(report, 'reports')

    part_name = chip.get('fpga', 'partname')
    dff_cells = []
    if chip.valid('fpga', part_name, 'var', 'vpr_registers'):
        dff_cells = chip.get('fpga', part_name, 'var', 'vpr_registers')
    brams_cells = []
    if chip.valid('fpga', part_name, 'var', 'vpr_brams'):
        brams_cells = chip.get('fpga', part_name, 'var', 'vpr_brams')
    dsps_cells = []
    if chip.valid('fpga', part_name, 'var', 'vpr_dsps'):
        dsps_cells = chip.get('fpga', part_name, 'var', 'vpr_dsps')

    stat_extract = re.compile(r'  \s*(.*)\s*:\s*([0-9]+)')
    lut_match = re.compile(r'([0-9]+)-LUT')
    route_length = re.compile(r'	Total wirelength: ([0-9]+)')
    log_file = f'{step}.log'
    mdata = {
        "registers": 0,
        "luts": 0,
        "dsps": 0,
        "brams": 0
    }
    with sc_open(log_file) as f:
        in_stats = False
        for line in f:
            if in_stats:
                if not line.startswith("  "):
                    in_stats = False
                    continue
                data = stat_extract.findall(line)
                if data:
                    dtype, value = data[0]
                    dtype = dtype.strip()
                    value = int(value)

                    if dtype == "Blocks":
                        record_metric(chip, step, index, "cells", value, log_file)
                    elif dtype == "Nets":
                        record_metric(chip, step, index, "nets", value, log_file)
                    elif dtype in dff_cells:
                        mdata["registers"] += value
                    elif dtype in dsps_cells:
                        mdata["dsps"] += value
                    elif dtype in brams_cells:
                        mdata["brams"] += value
                    else:
                        lut_type = lut_match.findall(dtype)
                        if lut_type:
                            if int(lut_type[0]) == 0:
                                pass
                            else:
                                mdata["luts"] += value
            else:
                if line.startswith("Circuit Statistics:"):
                    in_stats = True
                route_len_data = route_length.findall(line)
                if route_len_data:
                    # Fake the unit since this is meaningless for the FPGA
                    units = chip.get('metric', 'wirelength', field='unit')
                    record_metric(chip, step, index, 'wirelength',
                                  int(route_len_data[0]),
                                  log_file,
                                  source_unit=units)

    for metric, value in mdata.items():
        record_metric(chip, step, index, metric, value, log_file)

    if os.path.exists(__block_file):
        with sc_open(__block_file) as f:
            data = json.load(f)

            if "num_nets" in data and chip.get('metric', 'nets', step=step, index=index) is None:
                record_metric(chip, step, index, "nets", int(data["num_nets"]), __block_file)

            io = 0
            if "input_pins" in data:
                io += int(data["input_pins"])
            if "output_pins" in data:
                io += int(data["output_pins"])

            record_metric(chip, step, index, "pins", io, __block_file)

    for setup_report in ("reports/report_timing.setup.rpt",
                         "reports/pre_pack.report_timing.setup.rpt"):
        if not os.path.exists(setup_report):
            continue

        slack = _parse_timing_report(setup_report)
        if slack is not None:
            wns = min([slack, 0])
            record_metric(chip, step, index, "setupslack", slack, setup_report, source_unit="ns")
            record_metric(chip, step, index, "setupwns", wns, setup_report, source_unit="ns")
            break

    for hold_report in ("reports/report_timing.hold.rpt", ):
        if not os.path.exists(hold_report):
            continue

        slack = _parse_timing_report(hold_report)
        if slack is not None:
            wns = min([slack, 0])
            record_metric(chip, step, index, "holdslack", slack, hold_report, source_unit="ns")
            record_metric(chip, step, index, "holdwns", wns, hold_report, source_unit="ns")
            break

    unconstrained = None
    unconstrained_reports = []
    for unconstrained_report in ("reports/report_unconstrained_timing.hold.rpt",
                                 "reports/report_unconstrained_timing.setup.rpt"):
        if not os.path.exists(unconstrained_report):
            continue

        paths = _parse_unconstrained_report(unconstrained_report)
        if unconstrained is None:
            unconstrained = paths

        unconstrained = max([paths, unconstrained])
        unconstrained_reports.append(unconstrained_report)

    if unconstrained is not None:
        record_metric(chip, step, index, "unconstrained", unconstrained, unconstrained_reports)


def _parse_timing_report(report):
    slack = re.compile(r"slack \(.*\)\s+(-?\d+\.?\d*)")
    with sc_open(report) as f:
        for line in f:
            match_slack = slack.findall(line)
            if match_slack:
                return float(match_slack[0])
    return None


def _parse_unconstrained_report(report):
    path = re.compile(r"\d+ .*")
    count = 0
    with sc_open(report) as f:
        for line in f:
            if path.match(line):
                count += 1
    return count


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("vpr.json")



import siliconcompiler

from siliconcompiler.tools.openroad import rcx_bench
from siliconcompiler.tools.openroad import rcx_extract
from siliconcompiler.tools.builtin import nop
from siliconcompiler.flows._common import _make_docs


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    _make_docs(chip)
    return setup(corners=5)


###########################################################################
# Flowgraph Setup
############################################################################
def setup(extraction_task=None, corners=1, serial_extraction=False):
    '''
    Flow to generate the OpenRCX decks needed by OpenROAD to do parasitic
    extraction.
    '''

    flowname = 'generate_rcx'
    flow = siliconcompiler.Flow(flowname)

    if not extraction_task:
        flow.logger.warning('Valid extraction not specified, defaulting to builtin/nop')
        extraction_task = nop

    flow.node(flowname, 'bench', rcx_bench)
    prev = 'bench'
    for corner in range(corners):
        # For each corner generate a pex step to build the reference SPEF file
        # and the extract step to use the SPEF file to build the new OpenRCX deck
        flow.node(flowname, 'pex', extraction_task, index=corner)
        flow.node(flowname, 'extract', rcx_extract, index=corner)

        if corner == 0:
            prev = 'bench'
            prev_index = 0
        else:
            prev = 'pex'
            prev_index = corner - 1

        flow.edge(flowname, 'bench', 'pex', head_index=corner, tail_index=0)
        flow.edge(flowname, 'pex', 'extract', head_index=corner, tail_index=corner)
        flow.edge(flowname, 'bench', 'extract', head_index=corner, tail_index=0)

        if serial_extraction and corner > 0:
            # For license restrictions make each pex step dependent on the previous pex step
            flow.edge(flowname, prev, 'pex', head_index=corner, tail_index=prev_index)

    flow.node(flowname, 'bench', rcx_bench)

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



import glob
import os
import stat
from siliconcompiler.tools.execute.execute import setup as tool_setup
from siliconcompiler.tools._common import input_provides, get_tool_task


def setup(chip):
    '''
    Execute the output of the previous step directly.
    This only works if the task receives a single file.
    '''
    tool_setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'input',
             list(input_provides(chip, step, index).keys()),
             step=step, index=index)


def pre_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, _ = get_tool_task(chip, step, index)

    exec = None
    for fin in glob.glob('inputs/*'):
        if fin.endswith('.pkg.json'):
            continue
        exec = os.path.abspath(fin)
        break

    if not exec:
        chip.error(f'{step}{index} did not receive an executable file')

    chip.set('tool', tool, 'exe', exec)

    os.chmod(exec, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)



'''
OpenROAD is an automated physical design platform for
integrated circuit design with a complete set of features
needed to translate a synthesized netlist to a tapeout ready
GDSII.

Documentation: https://openroad.readthedocs.io/

Sources: https://github.com/The-OpenROAD-Project/OpenROAD

Installation: https://github.com/The-OpenROAD-Project/OpenROAD
'''
from siliconcompiler.tools._common import get_tool_task


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    from siliconcompiler.targets import asap7_demo
    chip.use(asap7_demo)


################################
# Setup Tool (pre executable)
################################
def setup(chip, exit=True, clobber=False):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', 'openroad')
    chip.set('tool', tool, 'vswitch', '-version')
    chip.set('tool', tool, 'version', '>=v2.0-17598', clobber=clobber)
    chip.set('tool', tool, 'format', 'tcl', clobber=clobber)

    option = [
        "-no_init",
        "-metrics", "reports/metrics.json"
    ]

    # exit automatically in batch mode and not breakpoint
    if exit and \
       not chip.get('option', 'breakpoint', step=step, index=index):
        option.append("-exit")

    chip.set('tool', tool, 'task', task, 'option', option,
             step=step, index=index, clobber=clobber)

    chip.set('tool', tool, 'task', task, 'refdir',
             'tools/openroad/scripts',
             step=step, index=index, package='siliconcompiler')

    # basic warning and error grep check on logfile
    chip.set('tool', tool, 'task', task, 'regex', 'warnings',
             r'^\[WARNING|Warning',
             step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'regex', 'errors',
             r'^\[ERROR',
             step=step, index=index, clobber=clobber)

    chip.set('tool', tool, 'task', task, 'var', 'debug_level',
             'list of "tool key level" to enable debugging of OpenROAD',
             field='help')

    if chip.get('option', 'nodisplay'):
        # Tells QT to use the offscreen platform if nodisplay is used
        chip.set('tool', tool, 'task', task, 'env',
                 'QT_QPA_PLATFORM', 'offscreen',
                 step=step, index=index)


################################
# Version Check
################################
def parse_version(stdout):
    # stdout will be in one of the following forms:
    # - 1 08de3b46c71e329a10aa4e753dcfeba2ddf54ddd
    # - 1 v2.0-880-gd1c7001ad
    # - v2.0-1862-g0d785bd84

    # strip off the "1" prefix if it's there
    version = stdout.split()[-1]

    pieces = version.split('-')
    if len(pieces) > 1:
        # strip off the hash in the new version style
        return '-'.join(pieces[:-1])
    else:
        return pieces[0]


def normalize_version(version):
    if '.' in version:
        return version.lstrip('v')
    else:
        return '0'


##################################################
if __name__ == "__main__":
    chip = make_docs()
    chip.write_manifest("openroad.json")



import os
from docutils import nodes
from sphinx.util.nodes import nested_parse_with_titles
from docutils.statemachine import ViewList
from sphinx.util.docutils import SphinxDirective

import siliconcompiler
from siliconcompiler.schema import Schema
from siliconcompiler.sphinx_ext.utils import (
    strong,
    code,
    para,
    keypath,
    build_table,
    build_section_with_target,
    build_list
)
from siliconcompiler.schema import utils
from siliconcompiler.sphinx_ext import sc_root as SC_ROOT


# Main Sphinx plugin
class SchemaGen(SphinxDirective):

    def run(self):
        self.env.note_dependency(
            os.path.join(SC_ROOT, 'siliconcompiler', 'schema', 'schema_cfg.py'))
        self.env.note_dependency(__file__)
        self.env.note_dependency(utils.__file__)

        schema = Schema().cfg

        return self.process_schema(schema)

    def process_schema(self, schema, parents=[]):
        if 'help' in schema:
            entries = [[strong('Description'), para(schema['shorthelp'])],
                       [strong('Type'), para(schema['type'])]]

            if schema['pernode'] != utils.PerNode.NEVER:
                entries.append([strong('Per step/index'), para(schema['pernode'])])

            entries.append([strong('Scope'), para(schema['scope'])])

            if 'enum' in schema['type']:
                entries.append([strong('Allowed Values'),
                                build_list([code(val) for val in schema['enum']])])

            if 'unit' in schema:
                entries.append([strong('Unit'), para(schema['unit'])])

            defvalue = schema['node']['default']['default']['value']
            switch_list = [code(switch) for switch in schema['switch']]
            entries.extend([[strong('Default Value'), para(defvalue)],
                            [strong('CLI Switch'), build_list(switch_list)]])

            examples = {}
            for example in schema['example']:
                name, ex = example.split(':', 1)
                examples.setdefault(name, []).append(ex)

            for name, exs in examples.items():
                examples = [code(ex.strip()) for ex in exs]
                p = None
                for ex in examples:
                    if not p:
                        p = para("")
                    else:
                        p += para("")
                    p += ex
                entries.append([strong(f'Example ({name.upper()})'), p])

            table = build_table(entries, colwidths=[25, 75])
            body = self.parse_rst(utils.trim(schema['help']))

            return [table, body]
        else:
            sections = []
            for key in schema.keys():
                if key == 'default':
                    for n in self.process_schema(schema['default'], parents=parents):
                        sections.append(n)
                else:
                    if not parents and key in ('history', 'library'):
                        continue
                    section_key = 'param-' + '-'.join(parents + [key])
                    section = build_section_with_target(key, section_key, self.state.document)
                    for n in self.process_schema(schema[key], parents=parents + [key]):
                        section += n
                    sections.append(section)

            # Sort all sections alphabetically by title. We may also have nodes
            # in this list that aren't sections if  `schema` has a 'default'
            # entry that's a leaf. In this case, we sort this as an empty string
            # in order to put this node at the beginning of the list.
            return sorted(sections, key=lambda s: s[0][0] if isinstance(s, nodes.section) else '')

    def parse_rst(self, content):
        rst = ViewList()
        # use fake filename 'inline' for error # reporting
        for i, line in enumerate(content.split('\n')):
            rst.append(line, 'inline', i)
        body = nodes.paragraph()
        nested_parse_with_titles(self.state, rst, body)

        return body


class CategorySummary(SphinxDirective):

    option_spec = {'category': str}

    def run(self):
        self.env.note_dependency(__file__)
        category = self.options['category']

        # List of documentation objects to return.
        new_doc = []
        section = nodes.section(ids=[nodes.make_id(f'{category}_summary')])

        chip = siliconcompiler.Chip('<design>')

        table = [[strong('parameter'), strong('description')]]

        # Descend through defaults until we find the real items
        prefix = [category]
        while 'default' in chip.getdict(*prefix).keys():
            prefix.append('default')

        for item in chip.getkeys(*prefix):
            key = para('')
            key += keypath([*prefix, item], self.env.docname)
            if 'shorthelp' in chip.getkeys(*prefix, item):
                shorthelp = chip.get(*prefix, item, field='shorthelp')
                table.append([key, para(shorthelp)])
            else:
                table.append([key, para("Contains sub-tree of parameters. See Schema.")])
        section += build_table(table)
        new_doc += section

        return new_doc


class CategoryGroupTable(SphinxDirective):

    def count_keys(self, schema, *keypath):
        cfgs = schema.getdict(*keypath)
        count = 0
        for key, cfg in cfgs.items():
            if schema._is_leaf(cfg):
                count += 1
            else:
                count += self.count_keys(schema, *keypath, key)

        return count

    def run(self):
        self.env.note_dependency(__file__)

        desc = {
            "option": "Compilation options",
            "tool": "Individual tool settings",
            "flowgraph": "Execution flow definition",
            "pdk": "PDK related settings",
            "asic": "ASIC related settings",
            "fpga": "FPGA related settings",
            "checklist": "Checklist related settings",
            "constraint": "Design constraint settings",
            "metric": "Metric tracking",
            "record": "Compilation history tracking",
            "package": "Packaging manifest",
            "datasheet": "Design interface specifications",
            "schematic": "Schematic specifications",

            # Nothing to document
            "library": "",
            "history": "",
            "input": "",
            "output": "",
            "schemaversion": "",
            "design": "",
            "arg": "",
        }

        schema = Schema()

        # Check if all groups have desc
        for group in schema.getkeys():
            if group not in desc:
                raise ValueError(f"{group} not found in group descriptions")

        # Check if all groups have schema
        for group in desc.keys():
            if group not in schema.getkeys():
                raise ValueError(f"{group} not found in schema")

        table = [[strong('Group'), strong('Parameters'), strong('Description')]]

        total = 0
        for group in schema.getkeys():
            text = desc[group]
            if len(text) == 0:
                continue

            key = para('')
            key += keypath([group], self.env.docname)

            count = self.count_keys(schema, group)
            total += count

            table.append([key, para(f'{count}'), para(text)])

        table.append([strong('Total'), para(f'{total}'), para('')])

        return build_table(table)


def setup(app):
    app.add_directive('schemagen', SchemaGen)
    app.add_directive('schema_category_summary', CategorySummary)
    app.add_directive('schema_group_summary', CategoryGroupTable)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from siliconcompiler import Flow

from siliconcompiler.tools.openroad import rdlroute
from siliconcompiler.tools.klayout import export


def setup():
    '''
    A flow to perform RDL routing and generate a GDS
    '''
    flow = Flow('interposerflow')
    flow.node('interposerflow', 'rdlroute', rdlroute)
    flow.node('interposerflow', 'write_gds', export)

    flow.edge('interposerflow', 'rdlroute', 'write_gds')

    return flow



#!/usr/bin/env python3
# Based on https://github.com/The-OpenROAD-Project/OpenROAD-flow-scripts/blob/84545863573a90bca3612298361adb35bf39e692/flow/util/markDontUse.py  # noqa E501

import re
import gzip
import argparse  # argument parsing


def processLibertyFile(input_file, logger=None):
    # Read input file
    if logger:
        logger.info(f"Opening file for replace: {input_file}")
    if input_file.endswith(".gz") or input_file.endswith(".GZ"):
        f = gzip.open(input_file, 'rt', encoding="utf-8")
    else:
        f = open(input_file, encoding="utf-8")
    content = f.read().encode("ascii", "ignore").decode("ascii")
    f.close()

    # Yosys-abc throws an error if original_pin is found within the liberty file.
    # removing
    pattern = r"(.*original_pin.*)"
    replace = r"/* \1 */;"
    content, count = re.subn(pattern, replace, content)
    if logger:
        logger.info(f"Commented {count} lines containing \"original_pin\"")

    # Yosys, does not like properties that start with : !, without quotes
    pattern = r":\s+(!.*)\s+;"
    replace = r': "\1" ;'
    content, count = re.subn(pattern, replace, content)
    if logger:
        logger.info(f"Replaced malformed functions {count}")

    # Yosys-abc throws an error if the units are specified in 0.001pf, instead of 1ff
    pattern = r"capacitive_load_unit\s+\(0.001,pf\);"
    replace = "capacitive_load_unit (1,ff);"
    content, count = re.subn(pattern, replace, content)
    if logger:
        logger.info(f"Replaced capacitive load {count}")

    # Return new text
    return content


if __name__ == "__main__":
    # Parse and validate arguments
    # ==============================================================================
    parser = argparse.ArgumentParser(
        description='Replaces occurrences of cells in def or verilog files')
    parser.add_argument('--patterns', '-p', required=True,
                        help='List of search patterns')
    parser.add_argument('--inputFile', '-i', required=True,
                        help='Input File')
    parser.add_argument('--outputFile', '-o', required=True,
                        help='Output File')
    args = parser.parse_args()

    processLibertyFile(args.inputFile, args.outputFile, args.patterns)



from siliconcompiler.tools.openroad.rcx_bench import setup_task
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Convert extracted results into RCX.
    '''

    # Generic tool setup.
    setup_task(chip)

    design = chip.top()

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'var', 'corner',
             'Parasitic corner to generate RCX file for',
             field='help')
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['tool', tool, 'task', task, 'var', 'corner']),
             step=step, index=index)

    if chip.valid('tool', tool, 'task', task, 'var', 'corner') and \
       chip.get('tool', tool, 'task', task, 'var', 'corner', step=step, index=index):
        corner = chip.get('tool', tool, 'task', task, 'var', 'corner', step=step, index=index)[0]
    else:
        # Placeholder since require will cause this to fail
        corner = 'corner'

    chip.add('tool', tool, 'task', task, 'input', f'{design}.def', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'input', f'{design}.{corner}.spef', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', f'{design}.{corner}.rcx', step=step, index=index)




'''
Icarus is a verilog simulator with full support for Verilog
IEEE-1364. Icarus can simulate synthesizable as well as
behavioral Verilog.

Documentation: https://steveicarus.github.io/iverilog/

Sources: https://github.com/steveicarus/iverilog

Installation: https://github.com/steveicarus/iverilog
'''


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    from siliconcompiler.tools.icarus.compile import setup
    setup(chip)
    return chip


################################
# Version Check
################################
def parse_version(stdout):
    # First line: Icarus Verilog version 10.1 (stable) ()
    return stdout.split()[3]


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("icarus.json")



import requests
import shutil
import tarfile
import zipfile

import os.path

from fasteners import InterProcessLock
from io import BytesIO
from urllib.parse import urlparse

from siliconcompiler import SiliconCompilerError
from siliconcompiler.package import get_download_cache_path
from siliconcompiler.package import aquire_data_lock, release_data_lock


def get_resolver(url):
    if url.scheme in ("http", "https"):
        return http_resolver

    return None


def http_resolver(chip, package, path, ref, url):
    data_path, data_path_lock = get_download_cache_path(chip, package, ref)

    if os.path.exists(data_path):
        return data_path, False

    # Acquire lock
    data_lock = InterProcessLock(data_path_lock)
    aquire_data_lock(data_path, data_lock)

    extract_from_url(chip, package, path, ref, url, data_path)

    release_data_lock(data_lock)

    return data_path, True


def extract_from_url(chip, package, path, ref, url, data_path):
    data_url = path
    headers = {}
    if os.environ.get('GIT_TOKEN') or url.username:
        headers['Authorization'] = f'token {os.environ.get("GIT_TOKEN") or url.username}'
    if "github" in data_url:
        headers['Accept'] = 'application/octet-stream'
    data_url = path
    if data_url.endswith('/'):
        data_url = f"{data_url}{ref}.tar.gz"
    chip.logger.info(f'Downloading {package} data from {data_url}')
    response = requests.get(data_url, stream=True, headers=headers)
    if not response.ok:
        raise SiliconCompilerError(f'Failed to download {package} data source.', chip=chip)

    fileobj = BytesIO(response.content)
    try:
        with tarfile.open(fileobj=fileobj, mode='r|gz') as tar_ref:
            tar_ref.extractall(path=data_path)
    except tarfile.ReadError:
        fileobj.seek(0)
        # Try as zip
        with zipfile.ZipFile(fileobj) as zip_ref:
            zip_ref.extractall(path=data_path)

    if 'github' in url.netloc and len(os.listdir(data_path)) == 1:
        # Github inserts one folder at the highest level of the tar file
        # this compensates for this behavior
        gh_url = urlparse(data_url)

        repo = gh_url.path.split('/')[2]

        gh_ref = gh_url.path.split('/')[-1]
        if repo.endswith('.git'):
            gh_ref = ref
        elif gh_ref.endswith('.tar.gz'):
            gh_ref = gh_ref[0:-7]
        elif gh_ref.endswith('.tgz'):
            gh_ref = gh_ref[0:-4]
        else:
            gh_ref = gh_ref.split('.')[0]

        if gh_ref.startswith('v'):
            gh_ref = gh_ref[1:]

        github_folder = f"{repo}-{gh_ref}"

        if github_folder in os.listdir(data_path):
            # This moves all files one level up
            git_path = os.path.join(data_path, github_folder)
            for data_file in os.listdir(git_path):
                shutil.move(os.path.join(git_path, data_file), data_path)
            os.removedirs(git_path)



import logging
import uuid
import math
from siliconcompiler import Chip
from siliconcompiler.optimizer import Optimizer
from siliconcompiler.flowgraph import _get_flowgraph_nodes

try:
    from vizier.service import clients as vz_clients
    from vizier.service import pyvizier as vz

    from jax import config
    config.update("jax_enable_x64", True)
    _has_vizier = True

    logging.getLogger('absl').setLevel(logging.CRITICAL)
    logging.getLogger('jax').setLevel(logging.CRITICAL)
except ModuleNotFoundError:
    _has_vizier = False


class VizierOptimizier(Optimizer):
    def __init__(self, chip):
        if not _has_vizier:
            raise RuntimeError("vizier is not available")

        super().__init__(chip)

        self.__problem = None
        self.__study = None

        self.__owner = chip.design
        self.__experiment_rounds = None
        self.__parallel_experiment = None

    def __init_parameters(self):
        search_space = self.__problem.search_space.root
        for name, info in self._parameters.items():
            values = info["values"]
            if info["type"] == 'float':
                search_space.add_float_param(name, values[0], values[1])
            elif info["type"] == 'int':
                search_space.add_int_param(name, values[0], values[1])
            elif info["type"] == 'bool':
                search_space.add_discrete_param(name, values)
            elif info["type"] == 'enum':
                if any([isinstance(v, str) for v in values]):
                    search_space.add_categorical_param(name, values)
                else:
                    search_space.add_discrete_param(name, values)

    def __init_goals(self):
        metric_information = self.__problem.metric_information

        for name, info in self._goals.items():

            vz_goal = None
            if info["goal"] == 'max':
                vz_goal = vz.ObjectiveMetricGoal.MAXIMIZE
            elif info["goal"] == 'min':
                vz_goal = vz.ObjectiveMetricGoal.MINIMIZE
            else:
                raise ValueError(f'{info["goal"]} is not a supported goal')

            metric_information.append(vz.MetricInformation(name, goal=vz_goal))

    def __init_study(self):
        # Setup client and begin optimization
        # Vizier Service will be implicitly created
        study_config = vz.StudyConfig.from_problem(self.__problem)
        study_config.algorithm = 'DEFAULT'
        if len(self._goals) > 1:
            if self.__parallel_experiment == 1:
                study_config.algorithm = 'GAUSSIAN_PROCESS_BANDIT'
            else:
                study_config.algorithm = 'QUASI_RANDOM_SEARCH'

        self.__study = vz_clients.Study.from_study_config(
            study_config,
            owner=self.__owner,
            study_id=uuid.uuid4().hex)

    def run(self, experiments=None, parallel=None):
        if not experiments:
            experiments = 10 * len(self._parameters)
            self._chip.logger.debug(f'Setting number of optimizer experiments to {experiments}')

        if not parallel:
            parallel = 1

        self.__parallel_experiment = parallel

        # Algorithm, search space, and metrics.
        self.__problem = vz.ProblemStatement()

        self._clear_results()

        self.__init_parameters()
        self.__init_goals()

        self.__init_study()

        self.__experiment_rounds = int(math.ceil(float(experiments) / parallel))
        accept = True
        try:
            for n in range(self.__experiment_rounds):
                if self.__run_round(n):
                    break
        except KeyboardInterrupt:
            pass
        except Exception as e:
            self._chip.logger.error(f"{e}")
            accept = False
        finally:
            if accept:
                self.__record_optimal()

            self.__study.delete()
            self.__study = None

    def __run_round(self, experiment_round):
        # create a new chip with a copy of its schema
        chip = Chip(self._chip.design)
        chip.schema = self._chip.schema.copy()

        suggestions = self.__setup_round(experiment_round, chip)

        # Start run
        try:
            chip.logger.info(
                f"Starting optimizer run ({experiment_round+1} / {self.__experiment_rounds})")
            chip.run()
        except KeyboardInterrupt:
            raise
        except Exception as e:
            chip.logger.error(f"{e}")

        return self.__record_round(chip, suggestions)

    def __setup_round(self, experiment_round, chip):
        org_flow = self._chip.get("option", "flow")
        org_jobname = self._chip.get("option", "jobname")

        jobname = f"{org_jobname}-{org_flow}-{experiment_round+1}"

        flow_map = {}

        if self.__parallel_experiment > 1:
            flow = f'optimize_{org_flow}'
            # Create new graph
            for m in range(self.__parallel_experiment):
                graph_name = f'opt{m+1}'
                flow_map[m] = {
                    "name": f'{jobname}/{graph_name}',
                    "prefix": f"{graph_name}.",
                    "suggestion": None
                }
                chip.graph(flow, org_flow, name=graph_name)

            # Complete nodes
            nodes = _get_flowgraph_nodes(chip, org_flow)
            for step, _ in list(nodes):
                nodes.append((step, None))
            nodes = set(nodes)

            # Forward node specific values
            for key in chip.schema.allkeys():
                if key[0] == 'history':
                    continue

                for value, step, index in chip.schema._getvals(*key):
                    node = (step, index)

                    if node in nodes:
                        for info in flow_map.values():
                            chip.set(
                                *key,
                                value,
                                step=f'{info["prefix"]}{step}',
                                index=index)
        else:
            flow = org_flow
            flow_map[0] = {
                "name": jobname,
                "prefix": "",
                "suggestion": None
            }

        # Setup each experiment
        for m, suggestion in enumerate(self.__study.suggest(count=self.__parallel_experiment)):
            self._chip.logger.info(f'Setting parameters for {flow_map[m]["name"]}')
            flow_map[m]["suggestion"] = suggestion

            for param_name, param_value in suggestion.parameters.items():
                self._set_parameter(
                    param_name,
                    param_value,
                    chip,
                    flow_prefix=flow_map[m]["prefix"])

        chip.set('option', 'jobname', jobname)
        chip.set('option', 'flow', flow)
        chip.set('option', 'quiet', True)

        steps = set()
        for info in list(self._goals.values()) + list(self._assertions.values()):
            for flow in flow_map.values():
                steps.add(f'{flow["prefix"]}{info["step"]}')
        chip.set('option', 'to', steps)

        return flow_map

    def __record_round(self, chip, suggestions):
        jobname = chip.get('option', 'jobname')

        # Record history
        self._chip.schema.cfg['history'][jobname] = chip.schema.history(jobname).cfg

        stop = False

        for trial_entry in suggestions.values():
            trial_suggestion = trial_entry['suggestion']

            measurement = {}
            self._chip.logger.info(f'Measuring {trial_entry["name"]}')
            for meas_name, meas_entry in self._goals.items():
                measurement[meas_name] = chip.get(
                    *meas_entry["key"],
                    step=f'{trial_entry["prefix"]}{meas_entry["step"]}',
                    index=meas_entry["index"])

                self._chip.logger.info(f'  Measured {meas_entry["print"]} = '
                                       f'{measurement[meas_name]}')

            failed = None
            if any([value is None for value in measurement.values()]):
                failed = "Did not record measurement goal"
            elif not self._check_assertions(chip, trial_entry["prefix"]):
                failed = "Failed to meet assertions"

            if failed:
                self._chip.logger.error(f'{trial_entry["name"]} failed: {failed}')
                trial_suggestion.complete(vz.Measurement(),
                                          infeasible_reason=failed)
            else:
                trial_suggestion.complete(vz.Measurement(measurement))
                stop |= self._check_stop_goal(measurement)

        return stop

    def __record_optimal(self):
        optimal_trials = list(self.__study.optimal_trials())
        for n, optimal_trial in enumerate(optimal_trials):
            optimal_trial = optimal_trial.materialize()

            self._add_result(
                optimal_trial.parameters,
                optimal_trial.final_measurement
            )



import os.path

import siliconcompiler
from siliconcompiler import __version__ as sc_version

from siliconcompiler.targets import get_targets
from siliconcompiler.flows import get_flows
from siliconcompiler.checklists import get_checklists
from siliconcompiler.libs import get_libs
from siliconcompiler.pdks import get_pdks
from siliconcompiler.apps import get_apps
from siliconcompiler.tools import get_tools


sc_root = os.path.dirname(os.path.dirname(os.path.abspath(siliconcompiler.__file__)))


def relpath(file):
    file = os.path.abspath(file)
    if file.startswith(sc_root):
        return os.path.relpath(file, sc_root)
    return None


def get_codeurl(file=None):
    base_url = f"https://github.com/siliconcompiler/siliconcompiler/blob/v{sc_version}"

    if not file:
        return base_url

    if os.path.isabs(file):
        file = relpath(file)
        if not file:
            return None

    return f"{base_url}/{file}"


def targets():
    modules = []
    for name, mod in get_targets().items():
        modules.append((mod, name))
    return modules


def flows():
    modules = []
    for name, mod in get_flows().items():
        modules.append((mod, name))
    return modules


def libraries():
    modules = []
    for name, mod in get_libs().items():
        modules.append((mod, name))
    return modules


def pdks():
    modules = []
    for name, mod in get_pdks().items():
        modules.append((mod, name))
    return modules


def tools():
    modules = []
    for name, mod in get_tools().items():
        modules.append((mod, name))
    return modules


def apps():
    modules = []
    for name, mod in get_apps().items():
        modules.append((mod, name))
    return modules


def checklists():
    modules = []
    for name, mod in get_checklists().items():
        modules.append((mod, name))
    return modules



from siliconcompiler.apps import sc_dashboard
from siliconcompiler.apps import sc_install
from siliconcompiler.apps import sc_issue
from siliconcompiler.apps import sc_remote
from siliconcompiler.apps import sc_server
from siliconcompiler.apps import sc_show
from siliconcompiler.apps import sc
from siliconcompiler.apps import smake


def get_apps():
    '''
    Returns a dict of builtin apps
    '''
    return {
        module.__name__.split(".")[-1]: module for module in (
            sc_dashboard,
            sc_install,
            sc_issue,
            sc_remote,
            sc_server,
            sc_show,
            sc,
            smake
        )
    }
